# -*- coding: utf-8 -*-
import os

from testtools import TestCase
from . import makeprefs, dummykey, _temp_home
from shutil import copy
from mock import patch, Mock
from contextlib import nested


class CacheTest(TestCase):
    def setUp(self):
        super(CacheTest, self).setUp()
        self.prefs = makeprefs()
        self.home = os.path.join('t', 'data', 'home')

    def tearDown(self):
        super(CacheTest, self).tearDown()

    def _makeit(self, *args, **kwargs):
        from lacli.cache import Cache
        return Cache(*args, **kwargs)

    def test_cache(self):
        assert self._makeit("")

    def test_prepare(self):
        with _temp_home() as home:
            cache = self._makeit(home)
            cache.prepare('foo', os.path.join('t', 'data', 'arc1'))
            archives = cache._for_adf('archives')
            self.assertEqual(len(archives), 1)
            self.assertEqual('foo',
                             next(archives.itervalues())['archive'].title)

    def test_cache_dir(self):
        d = 'archives'
        self.assertTrue(
            os.path.isdir(self._makeit(self.home)._cache_dir(d)))
        with _temp_home() as home:
            cache = self._makeit(home)
            self.assertFalse(os.path.exists(cache._cache_dir(d)))
            self.assertTrue(os.path.isdir(cache._cache_dir(d, write=True)))

    def test_archive_open(self):
        open_mock = Mock(return_value=None)
        import lacli.cache as cache
        with patch.object(cache, 'open', open_mock, create=True):
            with _temp_home() as home:
                cache = self._makeit(home)
                dname = os.path.join(home, 'archives')
                fname = os.path.join(dname, 'foo')
                cache._archive_open('foo', 'w')
                open_mock.assert_called_with(fname, 'w')
                self.assertTrue(os.path.isdir(dname))
                cache._cert_open('foo', 'w')
                dname = os.path.join(home, 'certs')
                fname = os.path.join(dname, 'foo')
                open_mock.assert_called_with(fname, 'w')
                self.assertTrue(os.path.isdir(dname))

    def test_certs(self):
        with _temp_home() as home:
            cache = self._makeit(home)
            self.assertEqual({}, cache.certs())
            cdir = os.path.join(home, 'certs')
            os.makedirs(cdir)
            certs = cache.certs()
            self.assertEqual(0, len(certs))
            copy(os.path.join(self.home, 'archives', 'sample.adf'), cdir)
            certs = cache.certs()
            self.assertEqual(1, len(certs))
            self.assertIn('12-345', certs)
            self.assertIn('archive', certs['12-345'])
            self.assertEqual('My 2013 vacation',
                             certs['12-345']['archive'].title)
            c = certs['12-345']['cert'].keys[1]
            self.assertTrue(hasattr(c, 'key'))
            self.assertTrue(hasattr(c, 'method'))
            self.assertTrue(hasattr(c, 'input'))
            self.assertEqual(dummykey, c.input)
            self.assertEqual(1, c.key)
            self.assertEqual('pbkdf2', c.method)

    def test_save_cert(self):
        import lacli.cache
        from lacli.adf import Archive, Meta, Signature
        from StringIO import StringIO
        with nested(
                patch.object(lacli.cache, 'NamedTemporaryFile', create=True),
                patch.object(lacli.cache, 'archive_slug', create=True),
                ) as (mock_open, slug):
            mock_open.return_value.__enter__.return_value = StringIO()
            meta = Meta('zip', 'xor', created='now')
            archive = Archive('foo', meta)
            slug.return_value = 'foo'
            cache = self._makeit(self.home)
            cache.save_cert({'archive': archive,
                             'signature': Signature(aid="foo",
                                                    uri="http://baz.com",
                                                    created="now")})
            args, kwargs = mock_open.call_args
            self.assertIn('prefix', kwargs)
            self.assertEqual('foo', kwargs['prefix'])
            adf = mock_open.return_value.__enter__.return_value.getvalue()
            self.assertEqual(ADF_EXAMPLE_1, adf)

    def test_import_cert(self):
        import lacli.cache
        from StringIO import StringIO
        with nested(
                patch.object(lacli.cache, 'NamedTemporaryFile', create=True),
                patch.object(lacli.cache, 'archive_slug', create=True),
                _temp_home()
                ) as (mock_open, slug, home):
            mock_open.return_value.__enter__.return_value = StringIO()
            slug.return_value = 'foo'
            cache = self._makeit(home)
            cert = os.path.join('t', 'data', 'longaccess-74-5N93.html')
            aid = cache.import_cert(cert)
            args, kwargs = mock_open.call_args
            self.assertIn('prefix', kwargs)
            self.assertEqual('foo', kwargs['prefix'])
            self.assertEqual('74-5N93', aid)

    def test_upload_complete(self):
        import lacli.cache
        cache = self._makeit(self.home)
        with nested(
                patch.object(lacli.cache, 'open', create=True),
                patch.object(lacli.cache, 'load_archive', create=True),
                patch.object(lacli.cache, 'make_adf', create=True)
                ) as (mock_open, mock_load, mock_adf):
            mock_load.return_value = {}
            ds = cache.upload_complete("foo", {'archive_key': 'bar'})
            self.assertIn('signature', ds)
            self.assertEqual('bar', ds['signature'].aid)
            self.assertEqual('http://longaccess.com/a/', ds['signature'].uri)
            args, kwargs = mock_adf.call_args
            self.assertEqual('bar', args[0][0].aid)


ADF_EXAMPLE_1 = """!archive
meta: !meta {cipher: xor, created: now, format: zip}
title: foo
--- !signature {aid: foo, created: now, uri: 'http://baz.com'}
"""



from __future__ import unicode_literals
import platform

try:
    from lacli.version import __version__
except ImportError:
    __version__ = "0.2.3"

def get_client_info(terse=False):
    el = ["Longaccess client"]
    el.append(__version__)
    el.append(platform.platform(True, terse))
    el.append("py"+platform.python_version())
    el.append(platform.machine())
    return " ".join(el)
# vim: et:sw=4:ts=4



import sys
import signal
import cmd
from abc import ABCMeta, abstractmethod


class LaBaseCommand(cmd.Cmd, object):
    __metaclass__ = ABCMeta

    prompt = 'lacli> '

    def __init__(self, registry, *args, **kwargs):
        if hasattr(signal, 'SIGBREAK'):
            signal.signal(signal.SIGBREAK, signal.default_int_handler)
        cmd.Cmd.__init__(self, *args, **kwargs)
        self.registry = registry

    @property
    def verbose(self):
        return self.registry.prefs['command']['verbose']

    @property
    def safe(self):
        return not self.registry.prefs['command']['unsafe']

    @property
    def batch(self):
        return self.registry.prefs['command']['batch']

    @batch.setter
    def batch(self, newvalue):
        self.registry.prefs['command']['batch'] = newvalue

    @property
    def cache(self):
        return self.registry.cache

    @property
    def debug(self):
        return self.registry.prefs['command']['debug']

    @property
    def prefs(self):
        return self.registry.prefs

    @property
    def session(self):
        return self.registry.session

    @session.setter
    def session(self, newsession):
        self.registry.session = newsession

    @abstractmethod
    def makecmd(self, options):
        """ Parse docopt style options and return a cmd-style line """

    def do_EOF(self, line):
        print
        return True

    def input(self, prompt=""):
        sys.stdout.write(prompt)
        line = sys.stdin.readline()
        if line:
            return line.strip()





import os
import sys
import hashlib

from tempfile import NamedTemporaryFile
from lacli.adf import Auth, make_adf
from lacli.crypt import CryptIO
from lacli.cipher import get_cipher
from lacli.hash import HashIO
from lacli.enc import get_unicode
from shutil import copyfileobj
from zipfile import ZipFile, ZIP_DEFLATED
from itertools import imap
from contextlib import contextmanager
try:
    import zipstream
except ImportError:
    zipstream = None


def archive_handle(docs):
    h = HashIO()
    make_adf(docs, out=h)
    return h.getvalue().encode('hex')


def restore_archive(archive, path, cert, folder, tmpdir, cb=None):
    cipher = get_cipher(archive, cert)
    with open(path, 'rb') as infile:
        with NamedTemporaryFile() as dst:
            with CryptIO(infile, cipher) as cf:
                copyfileobj(cf, dst)
            dst.flush()
            dst.seek(0)
            with ZipFile(dst) as zf:
                map(cb,
                    map(lambda zi: zf.extract(zi, folder), zf.infolist()))


class MyHashObj(object):
    hashf = None

    def __init__(self, hashf='sha512'):
        self.md5 = hashlib.md5()
        if hasattr(hashlib, hashf):
            self.hashf = hashf
            setattr(self, hashf, getattr(hashlib, hashf)())

    def update(self, data):
        self.md5.update(data)
        if self.hashf:
            getattr(self, self.hashf).update(data)

    def auth(self):
        args = {'md5': self.md5.digest()}
        if self.hashf:
            args[self.hashf] = getattr(self, self.hashf).digest()
        return Auth(**args)


def dump_archive(archive, items, cert, cb=None, tmpdir='/tmp',
                 hashf='sha512'):
    name = archive_handle([archive, cert])
    cipher = get_cipher(archive, cert)
    hashobj = MyHashObj(hashf)

    if sys.platform.startswith('win'):
        # windows has unicode file system api
        items = map(unicode, items)

    path = _writer(name, items, cb,
                   cipher, tmpdir, hashobj)

    return (name, path, hashobj.auth())


@contextmanager
def _temp_file(fdst, name, tmpdir):
    with NamedTemporaryFile(prefix=name, dir=tmpdir) as zf:
        yield zf
        zf.flush()
        zf.seek(0)
        with fdst as dst:
            copyfileobj(zf, dst, 1024)


def walk_folders(folders):
    for folder in folders:
        if not os.path.isdir(folder):
            yield (folder, os.path.basename(folder))
        else:
            for root, _, fs in os.walk(folder):
                for f in fs:
                    path = os.path.join(root, f)
                    strip = os.path.dirname(folder)
                    rel = os.path.relpath(path, strip)
                    yield (path, get_unicode(rel))


def _writer(name, items, cb, cipher, tmpdir, hashobj=None):

    tmpargs = {'delete': False,
               'suffix': ".longaccess",
               'dir': tmpdir,
               'prefix': name}
    dst = NamedTemporaryFile(**tmpargs)

    fdst = CryptIO(dst, cipher, hashobj=hashobj)

    if zipstream is None:
        # do it in two passes now as vanilla zipfile
        # can't easily handle streaming
        fdst = _temp_file(fdst, name, tmpdir)

    files = walk_folders(imap(os.path.abspath, items))

    def _args():
        for path, rel in files:
            try:
                cb(path, rel)
                yield ((path,), {'arcname': rel.encode('utf-8')})
            except Exception as e:
                if not hasattr(e, 'filename'):
                    setattr(e, 'filename', path)
                raise

    try:
        with fdst as zf:
            zipargs = {'mode': 'w', 'compression': ZIP_DEFLATED,
                       'allowZip64': True}
            if zipstream is None:
                with ZipFile(zf, **zipargs) as zpf:
                    for args, kwargs in _args():
                        zpf.write(*args, **kwargs)
            else:
                with zipstream.ZipFile(**zipargs) as zpf:
                    zpf.paths_to_write = _args()
                    for chunk in zpf:
                        zf.write(chunk)
    except Exception:
        path = dst.name
        dst.close()
        if os.path.exists(path):
            os.unlink(path)  # don't leave trash
        raise
    return dst.name



from lacli import __version__
from lacli.decorators import command
from lacli.log import getLogger
from lacli.compose import compose
from lacli.adf import make_adf, Archive, Certificate, Meta, Cipher
from lacli.archive import restore_archive
from lacli.basecmd import LaBaseCommand
from lacli.decorators import contains, login_async, expand_args
from lacli.exceptions import PauseEvent, UploadError
from twisted.python.log import msg, err, PythonLoggingObserver
from twisted.internet import reactor, defer, threads, task
from thrift.transport import TTwisted
from thrift.protocol import TBinaryProtocol
from lacli.server.interface.ClientInterface import CLI, ttypes
from lacli.server.error import tthrow, log_hide
from itertools import starmap
from lacli.progress import ServerProgressHandler
from lacli.upload import UploadState
from binascii import b2a_hex
from StringIO import StringIO
import os
import errno
import treq
import json


class LaServerCommand(LaBaseCommand, CLI.Processor):
    """Run a RPC server

    Usage: lacli server [--no-detach] [--port <port>] [--mhole]

    Options:
        --no-detach              don't detach from terminal
        --port <port>            port to listen on [default: 9090]
        --mhole                  start a manhole on the next port number
    """
    prompt = 'lacli:server> '

    def __init__(self, *args, **kwargs):
        super(LaServerCommand, self).__init__(*args, **kwargs)
        self.logincmd = self.registry.cmd.login
        self.manhole = False
        self.srm = self.prefs['command']['srm']
        if self.prefs['gui']['rememberme'] is True:
            self.logincmd.username = self.prefs['gui']['username']
            self.logincmd.password = self.prefs['gui']['password']
        UploadState.init(self.cache)
        CLI.Processor.__init__(self, self)

    def makecmd(self, options):
        cmd = ["run"]
        if options['--port']:
            cmd.append(options['--port'])
        if options['--mhole']:
            self.manhole = True
        return " ".join(cmd)

    def get_server(self):
        factory = TBinaryProtocol.TBinaryProtocolFactory()
        return TTwisted.ThriftServerFactory(
            processor=self, iprot_factory=factory)


    @command(port=int)
    def do_run(self, port=9090):
        """
        Usage: run [<port>]
        """
        reactor.listenTCP(port, self.get_server(), interface='127.0.0.1')
        if self.manhole is True:
            from twisted.manhole.telnet import ShellFactory
            f = ShellFactory()
            f.username = f.password = 'admin'
            reactor.listenTCP(port+1, f)
        tlog = PythonLoggingObserver()
        tlog.start()
        msg('Running reactor')
        self.batch = True
        if self.prefs['gui']['rememberme'] is True:
            self.logincmd.username = self.prefs['gui']['username']
            self.logincmd.password = self.prefs['gui']['password']
            self.logincmd.email = self.prefs['gui']['email']
        reactor.run()
        self.batch = False
        tlog.stop()

    def process(self, iprot, oprot):
        d = CLI.Processor.process(self, iprot, oprot)
        d.addErrback(self.logUnhandledError)
        return d

    def logUnhandledError(self, error):
        err(error)
        getLogger().debug("unhandled error: ", exc_info=True)
        return error

    def toArchiveInfo(self, docs):
        title = docs['archive'].title
        size = docs['archive'].meta.size
        created = docs['archive'].meta.created
        description = docs['archive'].description
        md5 = docs['auth'].md5.encode('hex')
        if title is not None:
            title = title.encode('utf8')
        if description is not None:
            description = description.encode('utf8')
        return ttypes.ArchiveInfo(
            title,
            description,
            size,
            ttypes.DateInfo(
                created.day,
                created.month,
                created.year,
                created.hour,
                created.minute,
                created.second),
            md5)

    def toArchive(self, fname, docs):
        status = self.cache.archive_status(fname, docs)
        ident = os.path.basename(fname)
        return ttypes.Archive(
            ident,
            status,
            self.toArchiveInfo(docs))

    def PingCLI(self):
        msg('pingCLI()')
        return True


    @tthrow
    def UserIsLoggedIn(self):
        if self.session is None:
            return False
        if self.logincmd.email is None:
            if self.logincmd.username is not None:
                try:
                    d = self.logincmd.login_async(
                        self.logincmd.username, self.logincmd.password)
                    d.addCallback(lambda x: True)
                    d.addErrback(lambda x: False)
                    return d
                except Exception as e:
                    getLogger().debug("Couldn't login", exc_info=True)
            return False
        else:
            return True
             
        

    @tthrow
    @log_hide(_args=True)
    @defer.inlineCallbacks
    def LoginUser(self, username, password, remember):
        yield self.logincmd.login_async(username, password)
        if remember:
            getLogger().debug("Saving credentials for {}".format(self.logincmd.username))
            self.prefs['gui']['username'] = self.logincmd.username
            self.prefs['gui']['password'] = self.logincmd.password
            self.prefs['gui']['rememberme'] = remember
            self.cache.save_prefs(self.prefs)
        defer.returnValue(True)

    @tthrow
    def Logout(self):
        self.logincmd.logout_batch()
        if self.prefs['gui']['rememberme'] is True:
            # clear credentials
            self.prefs['gui']['username'] = None
            self.prefs['gui']['password'] = None
            self.prefs['gui']['rememberme'] = False
            self.cache.save_prefs(self.prefs)
        return True

    @defer.inlineCallbacks
    def capsules(self):
        cs = yield self.session.async_capsules()
        defer.returnValue([
            ttypes.Capsule(
                '', str(c['id']), c['resource_uri'],
                c['title'], '', ttypes.DateInfo(),
                c['size'], c['remaining'], [])
            for c in cs])

    @tthrow
    @login_async
    def GetCapsules(self):
        """
        """
        return self.capsules()

    @tthrow
    def CreateArchive(self, paths):
        """
        """
        paths = map(lambda p: p.decode('utf-8'), paths)
        def progress(path, rel):
            if not path:
                msg("Encrypting..")
            else:
                msg(rel.encode('utf8'))

        d = threads.deferToThread(self.cache.prepare,
            "_temp", paths, description="_temp", cb=progress)
        d.addCallback(expand_args(self.toArchive))
        return d

    @tthrow
    def GetUploads(self):
        """
          list<Archive> GetUploads(),
        """
        archives = self.cache._for_adf('archives')
        return list(starmap(self.toArchive, archives.iteritems()))

    def reset_upload(self, archive, times=3):
        try:
            return UploadState.reset(archive)
        except OSError as e:
            if e.errno == errno.EACCES and times > 0:
                return task.deferLater(reactor, 2,
                    self.reset_upload, archive, times-1)
            raise

    @defer.inlineCallbacks
    def _upload(self, f, d, s):
        size = d['archive'].meta.size
        acmd = self.registry.cmd.archive
        with s:
            with ServerProgressHandler(maxval=size, state=s) as progq:
                with self.cache._archive_open(f, 'w') as _archive:
                    make_adf(list(d.itervalues()), out=_archive)
                saved = yield acmd.upload_async(d, f, progq, s)
            status = yield acmd._poll_status_async(saved['link'])
            getLogger().debug("API returned status: {}".format(status))
            if status['status'] == "error":
                raise UploadError()
            elif status['status'] == "completed":
                fname = saved['fname']
                getLogger().debug("upload completed")
                self.cache.save_cert(self.cache.upload_complete(fname, status))
                self.reset_upload(f)

    @tthrow
    @login_async
    @defer.inlineCallbacks
    def UploadToCapsule(self, archive, capsule, title, description):
        """
          void UploadToCapsule(1: string ArchiveLocalID, 2: string CapsuleID,
            3: string title, 4: string description)
        """
        docs = self.cache.get_adf(archive)
        docs['archive'].title = title
        docs['archive'].description = description
        status = self.cache.archive_status(archive, docs)
        if status != ttypes.ArchiveStatus.Local:
            raise ValueError("Archive state invalid")
        size = docs['archive'].meta.size
        cs = yield self.session.async_capsules()
        for c in cs:
            if c['id'] == int(capsule):
                if c.get('remaining', 0) >= docs['archive'].meta.size:
                    capsule = c
                    break
                else:
                    raise ValueError("Capsule is not big enough")
        else:
            raise ValueError("Capsule not found")

        state = UploadState.get(archive, size, capsule)
        if state.exc is not None:
            raise ValueError("Archive has error")

        self._upload(archive, docs, state)

    @tthrow
    @login_async
    def ResumeUpload(self, archive):
        """
          void ResumeUpload(1: string ArchiveLocalID)
        """
        docs = self.cache.get_adf(archive)
        status = self.cache.archive_status(archive, docs)
        if status == ttypes.ArchiveStatus.Completed:
            raise ValueError("Archive already complete")
        if status == ttypes.ArchiveStatus.Local:
            raise ValueError("Archive upload has not started yet.")
        state = UploadState.get(archive)
        self._upload(archive, docs, state)

    @tthrow
    def QueryArchiveStatus(self, archive):
        """
          TransferStatus QueryArchiveStatus(1: string ArchiveLocalID)
        """
        docs = self.cache.get_adf(archive)
        status = self.cache.archive_status(archive, docs)
        eta = description = ''
        remaining = progress = 0

        getLogger().debug("cached status: {}".format(status))
        if status == ttypes.ArchiveStatus.Completed:
            progress = docs['archive'].meta.size
        elif UploadState.has_state(archive):
            state = UploadState.get(archive)
            if state.exc is not None:
                status = ttypes.ArchiveStatus.Failed
            if state.active():
                status = ttypes.ArchiveStatus.InProgress
            getLogger().debug("state status: {}".format(status))
            progress = state.progress
            remaining = state.size - progress
            if state in ServerProgressHandler.uploads:
                elapsed = ServerProgressHandler.uploads[state].seconds_elapsed
                if progress > 0:
                    eta = str(int(elapsed * state.size / progress - elapsed))
        else:
            remaining = docs['archive'].meta.size
        return ttypes.TransferStatus(description, eta, remaining, progress, status)

    @tthrow
    def PauseUpload(self, archive):
        """
          void PauseUpload(1: string ArchiveLocalID)
        """
        if not UploadState.has_state(archive):
            raise ValueError("archive not found")
        UploadState.states[archive].pause()

    @login_async
    def _delete_upload(self, archive):
        if UploadState.has_state(archive):
            UploadState.states[archive].pause()
            self.reset_upload(archive)

    def _delete_archive(self, archive):
        self._delete_upload(archive)
        docs = self.cache.get_adf(archive)
        fname = self.cache.shred_archive(archive, self.srm, True)
        assert not os.path.exists(fname), fname + " still exists!"
        path = self.cache.data_file(docs['links'])
        self.cache.shred_file(path, self.srm, True)

    @tthrow
    def CancelUpload(self, archive):
        """
          void CancelUpload(1: string ArchiveLocalID)
        """
        return threads.deferToThread(self._delete_archive, archive)

    def toSignature(self, docs):
        sig = docs['signature']
        return ttypes.Signature(
            sig.aid,
            ttypes.DateInfo(
                sig.created.day,
                sig.created.month,
                sig.created.year,
                sig.created.hour,
                sig.created.minute,
                sig.created.second),
            sig.creator,
            "")

    def toCertificate(self, ident, docs):
        return ttypes.Certificate(
            b2a_hex(docs['cert'].key).upper(),
            self.toSignature(docs),
            self.toArchiveInfo(docs),
            ident)

    @tthrow
    def GetCertificates(self):
        """
          list<Certificate> getCertificates(),
        """
        certs = self.cache._for_adf('certs')
        return list(starmap(self.toCertificate, certs.iteritems()))

    @tthrow
    @log_hide(_ret=True)
    def ExportCertificate(self, archive, fmt):
        """
          binary ExportCertificate(1: string ArchiveID,
            2: CertExportFormat format)
        """
        docs = self.cache.get_adf(archive, 'certs')
        out = StringIO()
        if fmt == ttypes.CertExportFormat.PDF:
            raise NotImplementedError("not implemented")
        elif fmt == ttypes.CertExportFormat.YAML:
            make_adf(list(docs.itervalues()), out=out, pretty=True)
            return out.getvalue() 
        elif fmt == ttypes.CertExportFormat.HTML:
            out.write(self.cache._printable_cert(docs))
            return out.getvalue() 
        else:
            raise NotImplementedError("Unknown format")
            

    @tthrow
    @log_hide(_args=True)
    def Decrypt(self, path, key, dest):
        """
          void Decrypt(1: string archivePath,2: string key,
            3: string destinationPath)
        """
        cert = Certificate(key.decode('hex'))
        archive = Archive("_temp", Meta('zip', Cipher('aes-256-ctr', 1)),
                               description="_temp")
        def _print(f):
            msg("Extracting {}".format(f))
        return threads.deferToThread(
            restore_archive, archive, path, cert, dest,
            self.cache._cache_dir('tmp', write=True), _print)

    @tthrow
    def CloseWhenPossible(self):
        reactor.callLater(1, reactor.stop)

    @tthrow
    @log_hide(_ret=True)
    def GetSettings(self):
        return ttypes.Settings(
            self.prefs['gui']['username'] or "",
            self.prefs['gui']['password'] or "",
            self.prefs['gui']['rememberme'],
            self.cache._cache_dir('archives'),
            self.cache._cache_dir('certs'))

    @tthrow
    @log_hide(_ret=True)
    def SetSettings(self, settings):
        if self.logincmd.email == self.logincmd.username:
            if settings.StoredUserName != self.logincmd.username:
                self.logincmd.email = settings.StoredUserName
        self.logincmd.username = settings.StoredUserName
        self.logincmd.password = settings.StoredPassword
        self.prefs['gui']['rememberme'] = settings.RememberMe
        self.prefs['gui']['username'] = settings.StoredUserName
        self.prefs['gui']['password'] = settings.StoredPassword
        self.cache.save_prefs(self.prefs)
        # TODO handle folder settings

    @tthrow
    @defer.inlineCallbacks
    def GetLatestVersion(self):
        try:
            r = yield treq.get("http://download.longaccess.com/latest.json")
            if r.code != 200:
                raise Exception("error getting latest version: {} {}".format(r.code, r.phrase))
            r = yield treq.content(r)
            vinfo = {k: v for k, v in json.loads(r).iteritems()
                     if k in ("version", "description", "uri")} 
        except Exception as e:
            getLogger().debug("couldn't get latest version", exc_info=True)
            vinfo = {"version": __version__}
        defer.returnValue(ttypes.VersionInfo(**vinfo))

    @tthrow
    def GetVersion(self):
        return ttypes.VersionInfo(version=__version__)

# vim: et:sw=4:ts=4



from behave_cli.expect.steps import *
from behave_cli.files.steps import *
from behave_cli.api.steps import *
from behave_cli.s3mock.steps import *
from behave_cli.netrc.steps import *
from behave_cli.s3mock import setup as s3setup
from behave_cli.s3mock import teardown as s3teardown
from behave_cli.s3mock.steps import s3bucket_named
from Crypto import Random


def mp_setup(ctx):
    buckets = ctx.buckets  # save buckets before setup resets them
    s3setup(ctx)
    if len(buckets):
        for bucket in buckets:
            s3bucket_named(ctx, bucket)
    Random.atfork()


def mp_teardown(ctx):
    s3teardown(ctx)



import os

from lacli.adf import make_adf, Archive, Meta, Links, Certificate
from behave import step
from tempfile import NamedTemporaryFile


@step(u'I have 1 available archive')
def one_archive(context):
    context.execute_steps(u'Given I have 1 available archive titled "foo"')


@step(u'I have 1 available archive titled "{title}"')
def one_archive_titled(context, title):
    d = os.path.join(context.environ['HOME'], "Longaccess/archives")
    if not os.path.isdir(d):
        os.makedirs(d)
    context.archive = NamedTemporaryFile(dir=d, suffix='.adf')
    context.archive.write(make_adf(Archive(title, Meta('zip', 'aes-256-ctr'))))
    context.archive.flush()


@step(u'the archive titled "{title}" has a link to a local copy')
def archive_copy(context, title):
    assert context.archive
    context.archive.seek(0)
    context.archive.write(make_adf([
        Archive(title, Meta('zip', 'aes-256-ctr')),
        Links(local='file://' + os.path.join(context.environ['HOME'],
                                             "Longaccess/data/test"))]))
    context.archive.flush()


@step(u'the local copy for "{title}" is an empty file')
def archive_copy_empty(context, title):
    assert context.archive
    datadir = os.path.join(context.environ['HOME'], "Longaccess/data")
    if not os.path.exists(datadir):
        os.makedirs(datadir)
    with open(os.path.join(datadir, 'test'), 'w'):
        pass


@step(u'I have a certificate for the archive with title "{title}"')
def archive_cert(context, title):
    assert context.archive
    d = os.path.join(context.environ['HOME'], "Longaccess/certs")
    if not os.path.isdir(d):
        os.makedirs(d)
    context.cert = NamedTemporaryFile(dir=d, suffix='.adf')
    context.cert.write(make_adf([Archive(title, Meta('zip', 'aes-256-ctr')),
                                 Certificate()]))
    context.cert.flush()


@step(u'I have downloaded an archive containing "{folder}"')
def downloaded_archive(context, folder):
    prepare_archive_folder(context, folder)
    from glob import glob
    aglob = os.path.join(context.environ['HOME'], "Longaccess/archives/*")
    af = glob(aglob).pop()
    d = os.path.join(context.environ['HOME'], "Longaccess/certs")
    if not os.path.isdir(d):
        os.makedirs(d)
    from shutil import copy
    copy(af, d)


@step(u'I have {num} pending uploads')
def num_pending_uploads(context, num):
    for n in range(int(num)):
        pending_upload(context, "upload"+str(n))


@step(u'I have a pending upload titled "{title}"')
def pending_upload(context, title):
    prepare_archive(context, title)
    from glob import glob
    aglob = os.path.join(context.environ['HOME'], "Longaccess/archives/*")
    af = glob(aglob).pop()
    d = os.path.join(context.environ['HOME'], "Longaccess/uploads")
    if not os.path.isdir(d):
        os.makedirs(d)
    assert(context.mock_api)
    import urlparse
    url = urlparse.urljoin(context.mock_api.url(), 'path/to/api/upload/1')
    with open(af) as f:
        from lacli.adf import load_archive, make_adf
        docs = load_archive(f)
        docs['links'].upload = url
        with open(os.path.join(d, os.path.basename(af)), 'w') as out:
            make_adf(list(docs.itervalues()), out=out)


@step(u'the upload status is "{status}"')
def upload_status(context, status):
    if status == 'error':
        context.mock_api.test('uploadError', 'longaccessmock')
    elif status == 'completed':
        context.mock_api.test('uploadComplete', 'longaccessmock')


@step(u'there is an archive titled "{title}"')
def exists_archive_titled(context, title):
    context.execute_steps(u"""
        Given the command line arguments "archive create"
        When I run console script "lacli"
        Then I see ") {}"
        """.format(title))


@step(u'there is a completed certificate')
def exists_certificate(context):
    from glob import glob
    files = glob(os.path.join(context.environ['HOME'], "Longaccess/certs/*"))
    assert len(files) > 0, "there is a certificate"
    from lacli.adf import load_archive
    docs = {}
    with open(files[0]) as f:
        docs = load_archive(f)
    assert 'links' in docs
    assert 'archive' in docs
    assert hasattr(docs['links'], 'download')
    assert docs['links'].download == 'https://longaccess.com/yoyoyo'


@step(u'there are {num} pending uploads')
def pending_upload_num(context, num):
    from glob import glob
    aglob = os.path.join(context.environ['HOME'], "Longaccess/uploads/*")
    assert len(glob(aglob)) == int(num)


@step(u'I prepare an archive with a file "{title}"')
def prepare_archive(context, title):
    context.execute_steps(u'''
        Given an empty folder "{t}"
        And under "{{{t}}}" an empty file "{t}"
        And the command line arguments "archive create  -t "{t}" {{{t}}}"
        When I run console script "lacli"
        Then I see "archive prepared"'''.format(t=title))


@step(u'I prepare an archive with a directory "{title}"')
def prepare_archive_folder(context, title):
    context.execute_steps(u'''
        Given the command line arguments "archive create {title}"
        When I run console script "lacli"
        Then I see "archive prepared"'''.format(title=title))



# thanks to mathieu larose: http://mathieularose.com/function-composition-in-python


import functools

def compose(*functions):
    def compose2(f, g):
        return lambda x: f(g(x))
    return functools.reduce(compose2, functions)



import os

from behave import step
from urlparse import urlparse


@step(u'I have 1 capsule')
def one_capsule(context):
    context.mock_api.test('oneCapsule', 'longaccessmock')


@step(u'I store my credentials in "{file}"')
def my_store_in_netrc(context, file):
    assert(context.mock_api)
    p = urlparse(context.mock_api.url())
    context.execute_steps(u"""
        Given I store my credentials for "{}" in "{}"
        """.format(p.hostname, file))


@step(u'the Longaccess directory exists in HOME')
def make_longaccess(context):
    d = os.path.join(context.environ['HOME'], "Longaccess")
    if not os.path.isdir(d):
        os.makedirs(d)



"""
Usage: lacli [--help] [-u <user>] [-p <pass>] [--verbose]
             [--home <home>] [--debug <level>] [--batch]
             <command> [<args>...]
       lacli -i

Commands (run lacli <command> -h for options):

    archive         manage archives
    capsule         manage capsules
    certificate     manage certificates
    login           configure access credentials
    server          run a RPC server

Options:
    -i, --interactive              interactive mode
    -u <user>, --user <user>       user name
    -p <pass>, --password <pass>   user password
    -d <level>, --debug <level>    debug level, from 0 to 2 [default: 0]
    --home <home>                  conf/cache dir [default: {home}]
    -v, --verbose                  print verbose information
    --batch                        be brief, don't ask questions
    -h, --help                     print this help
"""

import sys
import os
import cmd

from lacli.log import setupLogging, getLogger
from docopt import docopt, DocoptExit
from lacli.command import LaCapsuleCommand, LaCertsCommand, LaArchiveCommand
from lacli.login import LaLoginCommand
from lacli.server import LaServerCommand
from lacli import get_client_info, __version__
from lacli.api import RequestsFactory
from lacli.cache import Cache
from lacli.registry import LaRegistry
from datetime import datetime


default_home = os.path.expanduser(os.path.join('~', 'Longaccess'))


def settings(options):
    """
        >>> prefs, cache = settings({'--home': '/tmp', '--debug': 2})
        >>> cache.home
        '/tmp'
        >>> prefs['command']['debug']
        2
        >>> prefs['api']['verify']
        True
    """
    try:
        debug = int(options.get('--debug', 0))
    except ValueError:
        print "error: illegal value for 'debug' parameter."
        raise

    verify = True
    if '0' == os.getenv('LA_API_VERIFY'):
        verify = False

    batch = options.get('--batch')
    if not batch and os.getenv('LA_BATCH_OPERATION'):
        batch = True

    unsafe = False
    if '0' == os.getenv('LA_SAFE'):
        unsafe = True

    prefs = {
        'api': {
            'user': options.get('--user'),
            'pass': options.get('--password'),
            'url': os.getenv('LA_API_URL'),
            'verify': verify,
            'factory': RequestsFactory
        },
        'command': {
            'debug': debug,
            'verbose': options.get('--verbose'),
            'batch': batch,
            'unsafe': unsafe,
            'srm': None
        },
        'gui': {
            'rememberme': False,
            'username': None,
            'password': None,
            'email': None
        }
    }
    if options.get('<command>'):
        prefs[options['<command>']] = options.get('<args>')

    home = options.get('--home', default_home)
    if not home or not os.path.isdir(home):
        if not batch:
            while not home or not os.path.isdir(home):
                if not home:
                    print "Enter directory for Longaccess data? [{}]: ".format(
                        default_home)
                    home = sys.stdin.readline().strip() or default_home
                elif os.path.isfile(home):
                    print "{} is not a directory.".format(home)
                    print "Enter directory for Longaccess data? : "
                    home = sys.stdin.readline().strip()
                else:
                    print home, "does not exist."
                    print "Should I create it? (yes/no): "
                    if sys.stdin.readline().strip().lower() != 'yes':
                        sys.exit('Unable to proceed without home directory')
                os.makedirs(home)

    cache = Cache(None)
    try:
        if not os.path.isdir(home):
            os.makedirs(home)
        cache = Cache(home)
        prefs = cache.merge_prefs(prefs)
    except Exception:
        if debug > 0:
            getLogger().exception("initialization exception")

    return (prefs, cache)


class LaCommand(cmd.Cmd):
    prompt = 'lacli> '
    archive = None
    capsule = None
    certificate = None
    login = None

    def __init__(self, cache, prefs):
        cmd.Cmd.__init__(self)
        registry = LaRegistry(cache, prefs, self)
        self.archive = LaArchiveCommand(registry)
        self.capsule = LaCapsuleCommand(registry)
        self.certificate = LaCertsCommand(registry)
        self.login = LaLoginCommand(registry)
        self.server = LaServerCommand(registry)

    def do_EOF(self, line):
        print
        return True

    def dispatch(self, subcmd, options):
        options.insert(0, subcmd)
        try:
            subcmd = getattr(self, subcmd)
        except AttributeError:
            print "Unrecognized command:", subcmd
            print(__doc__)
            return
        try:
            line = subcmd.makecmd(docopt(subcmd.__doc__, options))
        except DocoptExit as e:
            print e
            return
        self.dispatch_one(subcmd, line)

    def dispatch_one(self, subcmd, line, interactive=False):
        if line:
            subcmd.onecmd(line)
        elif interactive:
            subcmd.cmdloop()

    def do_archive(self, line):
        self.dispatch_one(self.archive, line, True)

    def do_capsule(self, line):
        self.dispatch_one(self.capsule, line, True)

    def do_certificate(self, line):
        self.dispatch_one(self.certificate, line, True)

    def do_login(self, line):
        self.login.onecmd('login ' + line)

    def do_server(self, line):
        self.server.onecmd('server ' + line)


def main(args=sys.argv[1:]):
    """Main function called by `laput` command.
    """
    options = docopt(__doc__.format(home=default_home),
                     version='lacli {}'.format(__version__),
                     options_first=True)
    prefs, cache = settings(options)
    with setupLogging(prefs['command']['debug'], logfile=cache.log):
        getLogger().debug("{} starting on {}".format(
            get_client_info(), datetime.now().isoformat()))
        cli = LaCommand(cache, prefs)
        if options['<command>']:
            cli.dispatch(options['<command>'], options['<args>'])
        elif options['--interactive']:
            cli.cmdloop()

if __name__ == "__main__":
    main()



import lacli.main
import os
import sys
import multiprocessing
os.environ['REQUESTS_CA_BUNDLE'] = os.path.join(
    os.path.dirname(sys.executable), 'cacert.pem')
os.environ['PATH'] = ';'.join([
    os.path.dirname(sys.executable), os.environ['PATH']])
multiprocessing.freeze_support()
lacli.main.main()



from testtools import TestCase, ExpectedException
from testtools.matchers import Contains
from mock import Mock, patch
from itertools import repeat, izip
from . import makeprefs
from lacli.exceptions import (ApiAuthException, ApiErrorException,
                              ApiNoSessionError)
import json


class ApiTest(TestCase):
    def setUp(self):
        super(ApiTest, self).setUp()
        self.prefs = makeprefs()['api']

    def tearDown(self):
        super(ApiTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacli.api import Api
        return Api(*args, **kw)

    def _factory(self, prefs):
        from lacli.api import RequestsFactory
        return RequestsFactory(prefs)

    def _mocksessions(self, rsps):
        return Mock(new_session=Mock(return_value=Mock(**rsps)))

    def _mockresponse(self, json, method='get', **kwargs):
        mattr = {
            'json.side_effect': json,
            'raise_for_status': Mock(**kwargs),
        }
        return Mock(**mattr)

    def _json_request(self, url, mock_call):
        args, kwargs = mock_call.call_args
        headers = kwargs['headers']
        self.assertEqual('application/json', headers['content-type'])
        self.assertEqual(url, args[0])
        return kwargs['data']

    def test_factory(self):
        import lacli.api
        mock_netrc = Mock(hosts={'bla.com': ('a', 'b', 'c')})
        mock_construct = Mock(return_value=mock_netrc)
        with patch.object(lacli.api, 'netrc', mock_construct, create=True):
            self._factory(self.prefs)
            self._factory({})
            self._factory({'url': None})
            f = self._factory({'url': 'http://bla.com'})
            self.assertEqual(f.prefs['user'], 'a')
            self.assertEqual(f.prefs['pass'], 'c')
            s = f.new_session()
            self.assertEqual(s.auth, ('a', 'c'))
            self.assertEqual(s.verify, True)
            f = self._factory({'url': 'http://bla.com', 'verify': False})
            s = f.new_session()
            self.assertEqual(s.verify, False)

    def test_api(self):
        assert self._makeit(self.prefs, Mock())

    def test_api_nosession(self):
        self.assertRaises(ApiNoSessionError, self._makeit,
                          self.prefs,
                          Mock(new_session=Mock(side_effect=Exception())))

    def test_api_no_prefs(self):
        import lacli.api
        mock_netrc = Mock(hosts={'stage.longaccess.com': ('a', 'b', 'c')})
        mock_construct = Mock(return_value=mock_netrc)
        with patch.object(lacli.api, 'netrc', mock_construct, create=True):
            self._makeit({})
            self._makeit({}, self._factory({}))

    def test_api_root(self):
        r = json.loads(LA_ENDPOINTS_RESPONSE)
        s = self._mocksessions({'get.return_value': self._mockresponse([r])})
        api = self._makeit(self.prefs, sessions=s)
        self.assertEqual(r, api.root)

    def test_no_capsules(self):
        caps = json.loads(LA_CAPSULES_RESPONSE)
        caps['objects'] = []
        caps['meta']['total_count'] = 0
        r = self._mockresponse([json.loads(LA_ENDPOINTS_RESPONSE), caps])
        s = self._mocksessions({'get.return_value': r})
        api = self._makeit(self.prefs, sessions=s)
        capsules = list(api.capsules())
        self.assertEqual(len(capsules), 0)

    def test_capsules(self):
        j = map(json.loads, [LA_ENDPOINTS_RESPONSE, LA_CAPSULES_RESPONSE])
        s = self._mocksessions({'get.return_value': self._mockresponse(j)})
        api = self._makeit(self.prefs, sessions=s)
        capsules = list(api.capsules())
        self.assertEqual(len(capsules), 2)

    def test_unauthorized(self):
        from requests.exceptions import HTTPError
        exc = HTTPError(response=Mock(status_code=401))
        r = self._mockresponse({}, side_effect=exc)
        s = self._mocksessions({'get.return_value': r})
        api = self._makeit(self.prefs, sessions=s)
        with ExpectedException(ApiAuthException):
            list(api.capsules())

    def test_tokens_error(self):
        er = self._mockresponse([json.loads(LA_ENDPOINTS_RESPONSE)])
        caps = json.loads(LA_CAPSULES_RESPONSE)
        caps['objects'] = []
        caps['meta']['total_count'] = 0
        cr = self._mockresponse([caps])
        ur = self._mockresponse(repeat(json.loads(LA_UPLOAD_RESPONSE)))
        s = self._mocksessions({'get.side_effect': [er, cr],
                               'post.return_value': ur,
                               'patch.return_value': ur})
        api = self._makeit(self.prefs, sessions=s)
        tmgr = api.upload(1, Mock(title='', description=''))
        self.assertRaises(ValueError, tmgr.__enter__)

    def test_tokens(self):
        er = self._mockresponse([json.loads(LA_ENDPOINTS_RESPONSE)])
        cr = self._mockresponse([json.loads(LA_CAPSULES_RESPONSE)])
        ur = self._mockresponse(repeat(json.loads(LA_UPLOAD_RESPONSE)))
        post = Mock(return_value=ur)
        patch = Mock(return_value=ur)
        get = Mock(side_effect=[er, cr, ur, ur, ur, ur, Exception("foo")])
        s = self._mocksessions({'get': get, 'post': post, 'patch': patch})
        api = self._makeit(self.prefs, sessions=s)
        meta = Mock(size=1)
        auth = Mock(md5="aa".decode('hex'), sha512="bb".decode('hex'))
        tmgr = api.upload(1, Mock(title='', meta=meta, description=None), auth)
        url = 'http://baz.com/api/v1/upload/'
        with tmgr as upload:
            data = self._json_request(url, post)
            self.assertTrue('"description": ""' in data)
            for seq, token in izip(xrange(3), upload['tokens']):
                self.assertIn('token_access_key', token)
        data = self._json_request(url+"1/", patch)
        self.assertThat(data, Contains('"checksums": '))
        self.assertThat(data, Contains('"sha512": "bb"'))
        self.assertThat(data, Contains('"md5": "aa"'))
        self.assertThat(data, Contains('"size": 1'))
        status = api.upload_status(url+"1/")
        self.assertTrue('status' in status)
        self.assertEqual('pending', status['status'])
        self.assertRaises(ApiErrorException, api.upload_status, url+"1/")


LA_UPLOAD_RESPONSE = """{
    "id": 1,
    "capsule": "/api/v1/capsule/1/",
    "title": "foo",
    "description": "foo",
    "resource_uri": "/api/v1/upload/1/",
    "status": "pending",
    "token_access_key": "XXXXXXXXXXXXXXXXXXXX",
    "token_secret_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
    "token_session": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
    "token_expiration": "",
    "token_uid": "arn:aws:iam::XXXXXXXXXXXX:user/sts-test-dummy",
    "bucket": "lastage",
    "prefix": "prefix"
}
"""


LA_ENDPOINTS_RESPONSE = """{
  "account": {
     "list_endpoint": "/api/v1/account/",
     "schema": "/api/v1/account/schema/"
  },
  "archive": {
     "list_endpoint": "/api/v1/archive/",
     "schema": "/api/v1/archive/schema/"
  },
  "capsule": {
     "list_endpoint": "/api/v1/capsule/",
     "schema": "/api/v1/capsule/schema/"
  },
  "upload": {
     "list_endpoint": "/api/v1/upload/",
     "schema": "/api/v1/upload/schema/"
  }
}
"""

LA_CAPSULES_RESPONSE = """{
  "meta": {
     "limit": 20,
     "next": null,
     "offset": 0,
     "previous": null,
     "total_count": 2
  },
  "objects": [
     {
         "created": "2013-06-07T10:45:01",
         "id": 3,
         "resource_uri": "/api/v1/capsule/3/",
         "title": "Photos",
         "user": "/api/v1/user/3/"
      },
      {
          "created": "2013-06-07T10:44:38",
          "id": 2,
          "resource_uri":
          "/api/v1/capsule/2/",
          "title": "Stuff",
          "user": "/api/v1/user/2/"
       }
    ]
}"""



import os
import shutil

from tempfile import mkdtemp
from testtools import TestCase
from mock import patch
from StringIO import StringIO


class SettingsTest(TestCase):
    def setUp(self):
        super(SettingsTest, self).setUp()
        self.home = os.path.join('t', 'data', 'home')

    def tearDown(self):
        super(SettingsTest, self).tearDown()

    def _makeit(self):
        from lacli.main import settings
        return settings

    def test_home(self):
        settings = self._makeit()
        _, cache = settings({'--home': self.home})
        self.assertEqual(self.home, cache.home)

    @patch('lacli.main.sys.stdin', new_callable=StringIO)
    def test_make_no_home(self, stdin):
        try:
            settings = self._makeit()
            settings({
                '--batch': True,
                '--home': None
            })
        except SystemExit as e:
            self.assertEqual("None does not exist!", e.message)

        try:
            stdin.write("no\n")
            settings = self._makeit()
            settings({
                '--batch': True,
                '--home': None
            })
        except SystemExit as e:
            self.assertEqual("None does not exist!", e.message)

    @patch('lacli.main.sys.stdin', new_callable=StringIO)
    def test_make_home(self, stdin):
        try:
            tmpdir = mkdtemp()
            tmphome = os.path.join(tmpdir, 'a')
            stdin.write("yes\n")
            stdin.seek(0)
            settings = self._makeit()
            _, cache = settings({'--home': tmphome})
            self.assertEqual(tmphome, cache.home)
            self.assertTrue(os.path.isdir(tmphome))
        finally:
            if (tmpdir and os.path.isdir(tmphome)):
                shutil.rmtree(tmpdir)



# -*- coding: utf-8 -*-
import os

from testtools import TestCase
from . import makeprefs
from shutil import rmtree
from contextlib import contextmanager
from tempfile import mkdtemp
from mock import Mock


class ArchiveTest(TestCase):
    def setUp(self):
        super(ArchiveTest, self).setUp()
        self.prefs = makeprefs()
        self.home = os.path.join('t', 'data', 'home')

    def tearDown(self):
        super(ArchiveTest, self).tearDown()

    @contextmanager
    def _temp_home(self):
        d = mkdtemp()
        yield d
        rmtree(d)

    def test_slugify(self):
        from lacli.archive import _slugify
        self.assertEqual(_slugify("This is a test"), "this-is-a-test")
        self.assertEqual(_slugify(u"γειά σου ρε"), "geia-sou-re")

    def test_restore(self):
        from lacli.archive import restore_archive
        from lacli.cache import Cache
        cache = Cache(self.home)
        cert = cache.certs()['74-5N93']
        with self._temp_home() as tmpdir:
            cb = Mock()
            cfname = os.path.join(
                self.home, 'data', '2013-10-22-foobar.zip.crypt')
            restore_archive(
                cert['archive'], cfname, cert['cert'], tmpdir, tmpdir, cb)
            self.assertEqual(1, cb.call_count)
            self.assertTrue(os.path.exists(
                os.path.join(tmpdir, 'xtQz6ziJ.sh.part')))



import os

from testtools import TestCase
from . import makeprefs
from mock import Mock


class RegistryTest(TestCase):
    def setUp(self):
        super(RegistryTest, self).setUp()
        self.prefs = makeprefs(lambda x: None)
        self.home = os.path.join('t', 'data', 'home')

    def tearDown(self):
        super(RegistryTest, self).tearDown()

    def _makeit(self, *args, **kwargs):
        from lacli.registry import LaRegistry
        return LaRegistry(*args, **kwargs)

    def test_registry(self):
        assert self._makeit(Mock(), self.prefs, Mock())



import os

from urlparse import urlparse
from netrc import netrc
from lacli.log import getLogger
from lacli.decorators import cached_property


API_URL = 'https://www.longaccess.com/api/v1/'


class LaRegistry(object):
    cache = None
    prefs = None
    session = None

    def __init__(self, cache, prefs, cmd):
        if prefs['api'].get('url') is None:
            prefs['api']['url'] = API_URL
        self.cache = cache
        self.prefs = prefs
        self.cmd = cmd
        self.session = self.new_session()

    def init_prefs(self):
        prefs = self.prefs['api']
        if prefs.get('url') is None:
            prefs['url'] = API_URL
        if not prefs['user']:
            prefs['user'] = self._saved_session[0]
            prefs['pass'] = self._saved_session[1]
        return prefs

    def new_session(self, prefs=None):
        if not prefs:
            prefs = self.init_prefs()
        return self.prefs['api']['factory'](prefs)

    def netarsee(self):
        ours = os.path.join(self.cache.home, ".netrc")
        users = os.path.expanduser('~/.netrc')
        if not os.path.exists(ours) and os.path.exists(users):
            return users
        return ours

    @cached_property
    def _saved_session(self):
        hostname = urlparse(self.prefs['api']['url']).hostname
        try:
            for host, creds in netrc(self.netarsee()).hosts.iteritems():
                if host == hostname:
                    return (creds[0], creds[2])
        except:
            getLogger().debug("Couldn't read from netrc", exc_info=True)
        return (None, None)

    def save_session(self, *args):
        self._saved_session = tuple(args)
        with open(self.netarsee(), 'a') as f:
            f.write("machine {} login {} password {}\n".format(
                urlparse(self.session.url).hostname,
                self._saved_session[0],
                self._saved_session[1]))



import yaml
import pyaml
import sys
import json
import mmap
import re

from lacli.cipher import cipher_modes, new_key
from yaml import SafeLoader
from yaml import SafeDumper
from lacli.exceptions import InvalidArchiveError
from lacli.date import parse_timestamp, today, later, format_timestamp
from datetime import datetime
from base64 import b64encode, b64decode


class PrettySafeLoader(SafeLoader):
    pass


class BaseYAMLObject(yaml.YAMLObject):
    yaml_loader = PrettySafeLoader
    yaml_dumper = SafeDumper

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_mapping(cls.yaml_tag, data.__dict__)


class Archive(BaseYAMLObject):
    """
    >>> meta = Meta('zip', 'aes-256-ctr', size=1024, created='now')
    >>> archive = Archive('title', meta, tags=['foo', 'bar'])
    >>> archive.title
    'title'
    >>> archive.description
    >>> archive.meta.cipher
    'aes-256-ctr'
    >>> archive.meta.size
    1024
    >>> archive.meta.created
    'now'
    >>> archive.tags[1]
    'bar'
    >>> archive.description = 'what a wonderful dataset'
    >>> archive.description
    'what a wonderful dataset'
    >>> Archive('title', meta, description='worthless junk').description
    'worthless junk'
    >>> pyaml.dump(Archive('foo', meta), sys.stdout)
    !archive
    meta: !meta
      cipher: aes-256-ctr
      created: now
      format: zip
      size: 1024
    title: foo
    """
    yaml_tag = u'!archive'
    title = None
    description = None
    tags = None
    meta = None

    def __init__(self, title, meta, description=None, tags=[]):
        self.title = title
        if description:
            self.description = description
        if len(tags):
            self.tags = tags
        if not hasattr(meta, 'format'):
            raise ValueError("invalid meta: " + str(meta))
        self.meta = meta


class Auth(BaseYAMLObject):
    """
    >>> with open('t/data/home/certs/2013-10-13-foobar.adf') as f:
    ...     doc=load_archive(f)['auth']
    ...     auth = Auth(sha512=doc.sha512)
    ...     auth.sha512.encode('hex')[:10]
    'd5ad8f4cb5'
    """
    yaml_tag = u'!auth'

    def __init__(self, *args, **kwargs):
        self.__dict__.update(kwargs)


class Meta(BaseYAMLObject):
    """
    >>> meta = Meta('zip', 'aes-256-ctr', size=1024)
    >>> meta.format
    'zip'
    >>> meta.cipher
    'aes-256-ctr'
    >>> meta.size
    1024
    >>> meta = Meta('zip', 'aes-256-ctr')
    >>> meta.size
    >>> meta.size = 1024
    >>> meta.size
    1024
    """
    yaml_tag = u'!meta'
    size = None
    format = None
    cipher = None
    email = None
    name = None

    def __init__(self, format, cipher, size=None, created=None,
                 email=None, name=None):
        self.format = format
        self.cipher = cipher
        if size:
            self.size = size
        if email:
            self.email = email
        if name:
            self.name = name
        if created:
            self.created = created
        else:
            self.created = today()


class Format(BaseYAMLObject):
    yaml_tag = u'!format'


class Links(BaseYAMLObject):
    """
    >>> links = Links(local='http://foo.bar.com')
    >>> links.download
    >>> links.upload
    >>> links.local
    'http://foo.bar.com'
    >>> from StringIO import StringIO
    >>> out = StringIO()
    >>> make_adf(links, out=out)
    >>> print out.getvalue()
    !links {local: 'http://foo.bar.com'}
    <BLANKLINE>
    """
    yaml_tag = u'!links'
    download = None
    local = None
    upload = None
    """
    This attribute is undocumented in the ADF format, because it only makes
    sense in the context of a specific client. It contains the URL of the
    pending upload API resource.
    """

    def __init__(self, download=None, local=None, upload=None):
        if download:
            # deprecated, but for compatibility with early client
            self.download = download
        if local:
            self.local = local
        if upload:
            self.upload = upload


class Cipher(BaseYAMLObject):
    """
    >>> cipher = Cipher('aes-256-ctr', key=2, input='\1\2\3\4'*4)
    >>> cipher.mode
    'aes-256-ctr'
    >>> cipher.key
    2
    >>> cipher.input.encode('hex')
    '01020304010203040102030401020304'
    >>> Cipher('aes-256-ctr', key='foo')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "lacli/adf.py", line 70, in __init__
    ValueError: key must be integer
    >>> Cipher('wtf')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "lacli/adf.py", line 68, in __init__
    ValueError: wtf not recognized
    >>> cipher = Cipher('aes-256-ctr')
    >>> hasattr(cipher, 'key')
    False
    >>> hasattr(cipher, 'input')
    False
    >>> cipher = Cipher('aes-256-ctr', key=1)
    >>> hasattr(cipher, 'key')
    False
    """

    yaml_tag = u'!cipher'

    def __init__(self, mode, key=1, input=None):
        if mode not in cipher_modes():
            raise ValueError("{} not recognized".format(mode))
        if not isinstance(key, int):
            raise ValueError("key must be integer")
        self.mode = mode
        if key > 1:
            self.key = key
        if input:
            self.input = input

    def __setstate__(self, d):
        assert d['mode'] in cipher_modes()
        self.__dict__.update(d)


class DerivedKey(BaseYAMLObject):
    yaml_tag = u'!key'


class Certificate(BaseYAMLObject):
    """
    >>> cert = Certificate(chr(255)*16)
    >>> pyaml.dump(cert, sys.stdout)
    !certificate
    key: !!binary |
      /////////////////////w==
    """
    yaml_tag = u'!certificate'

    def __init__(self, key=None):
        if not key:
            self.key = new_key(256)
        else:
            self.key = key


class MAC(BaseYAMLObject):
    yaml_tag = u'!mac'


class Signature(BaseYAMLObject):
    """
    >>> sig = Signature('1', 'http://baz.com', 'now')
    >>> pyaml.dump(sig, sys.stdout)
    !signature
    aid: '1'
    created: now
    uri: http://baz.com
    """
    yaml_tag = u'!signature'
    aid = None
    uri = None
    created = None
    expires = None
    creator = None

    def __init__(self, aid, uri, created=None, expires=None, creator=None):
        self.aid = aid
        self.uri = uri
        if created:
            self.created = created
        else:
            self.created = today()

        if expires:
            self.expires = expires
        else:
            self.expires = later(self.created, years=30)

        if creator:
            self.creator = creator


def add_path_resolver(tag, keys):
    yaml.add_path_resolver(tag, keys, dict, Loader=PrettySafeLoader)

add_path_resolver(u'!format', ["meta", "format"])
add_path_resolver(u'!cipher', ["meta", "cipher"])
add_path_resolver(u'!meta', ["meta"])
add_path_resolver(u'!mac', ["mac"])
add_path_resolver(u'!signature', ["signature"])
add_path_resolver(u'!key', ["keys", None])

json_cert_re = re.compile(r"""
<template><script>var json_certificate='(.*)';</script></template>
""")


def load_archive(f):

    try:
        f = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
    except Exception as e:
        raise InvalidArchiveError(e)
    match = json_cert_re.search(f)
    d = {}
    if match:  # we have a json encoded cert in this file
        try:
            d = as_adf(match.group(1))
        except Exception as e:
            raise InvalidArchiveError(e)
    else:
        try:
            os = yaml.load_all(f, Loader=PrettySafeLoader,
                               tz_aware_datetimes=True)
        except TypeError:
            os = yaml.load_all(f, Loader=PrettySafeLoader)
        for o in os:
            if isinstance(o, Archive):
                d['archive'] = o
            elif isinstance(o, Auth):
                d['auth'] = o
            elif isinstance(o, Certificate):
                d['cert'] = o
            elif isinstance(o, Links):
                d['links'] = o
            elif isinstance(o, Signature):
                d['signature'] = o
    if 'links' in d and d['links'].download:
        # fix for early client saving archive id in links section
        assert 'signature' not in d
        d['signature'] = Signature(aid=d['links'].download,
                                   uri='http://longaccess.com/a/')
        if d['links'].upload or d['links'].local:
            d['links'].download = None
        else:
            del d['links']
    if 'archive' in d:
        return d
    raise InvalidArchiveError()


def make_adf(archive=None, out=None, pretty=False):
    """
    >>> meta = Meta('zip', 'aes-256-ctr', created='now')
    >>> archive = Archive('title', meta)
    >>> cert = Certificate(chr(255)*16)
    >>> make_adf([archive, cert], out=sys.stdout, pretty=True)
    !archive
    meta: !meta
      cipher: aes-256-ctr
      created: now
      format: zip
    title: title
    --- !certificate
    key: !!binary |
      /////////////////////w==
    >>> make_adf([archive, cert], out=sys.stdout)
    !archive
    meta: !meta {cipher: aes-256-ctr, created: now, format: zip}
    title: title
    --- !certificate
    key: !!binary |
      /////////////////////w==
    """

    if not hasattr(archive, '__getitem__'):
        archive = [archive]
    if pretty:
        out.write("--- ".join(map(pyaml.dump, archive)))
        return
    return yaml.safe_dump_all(
        archive, out, canonical=True, allow_unicode=True)


def archive_size(archive):
    size = ""
    if archive.meta.size:
        kib = archive.meta.size // 1024
        mib = kib // 1024
        gib = mib // 1024

        if not kib:
            size = "{} B".format(archive.meta.size)
        elif not mib:
            size = "{} KB".format(kib)
        elif not gib:
            size = "{} MB".format(mib)
        else:
            size = "{} GB".format(gib)
    return size


class ADFEncoder(json.JSONEncoder):
    def _b64(self, v):
        return {'_base64': b64encode(v)}

    def default(self, o):
        if isinstance(o, (list, dict, str, unicode,
                          int, float, bool, type(None))):
            return super(ADFEncoder, self).default(o)
        if isinstance(o, datetime):
            return format_timestamp(o)
        if o.yaml_tag == Certificate.yaml_tag:
            return {'key': self._b64(o.key)}
        if o.yaml_tag == Auth.yaml_tag:
            return dict([(k, self._b64(v)) for k, v in o.__dict__.iteritems()])
        return o.__dict__


def _as_adf_object(dct):
    """
    May throw parse error for dates
    """
    if '_base64' in dct:
        return b64decode(dct['_base64'])
    if 'key' in dct:
        return Certificate(key=dct['key'])
    if 'md5' in dct:
        return Auth(**dct)
    if 'mode' in dct:
        return Cipher(**dct)
    if 'meta' in dct:
        return Archive(**dct)
    if 'download' in dct or 'upload' in dct or 'local' in dct:
        return Links(**dct)
    if 'created' in dct:
        dct['created'] = parse_timestamp(dct['created'])
        if 'aid' in dct:
            return Signature(**dct)
        else:
            return Meta(**dct)
    return dct


def as_json(docs):
    return ADFEncoder(ensure_ascii=False).encode(docs)


def as_adf(data):
    return json.loads(data, object_hook=_as_adf_object)


def creation(docs):
    created = docs['archive'].meta.created
    sig = docs.get('signature')
    if sig and sig.created:
        created = sig.created
    return parse_timestamp(created)



from testtools import TestCase
from moto import mock_s3
from boto import connect_s3


class MPConnectionTest(TestCase):
    @classmethod
    def setup_class(cls):
        cls._token = {
            'token_access_key': '',
            'token_secret_key': '',
            'token_session': '',
            'token_expiration': '',
            'token_uid': '',
            'bucket': 'lastage',
            'prefix': 'upload/14/',
        }
        cls._bucket = 'lastage'

    def setUp(self):
        self.s3 = mock_s3()
        self.s3.start()
        boto = connect_s3()
        boto.create_bucket(self._bucket)

        super(MPConnectionTest, self).setUp()

    def tearDown(self):
        self.s3.stop()
        super(MPConnectionTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacli.pool import MPConnection
        return MPConnection(*args,  **kw)

    def test_mpconnection(self):
        assert self._makeit(self._token)

    def test_mpconnection_nouid(self):
        token = self._token
        token['token_uid'] = None
        assert self._makeit(token)

    def test_getconnection(self):
        conn = self._makeit(self._token)
        assert conn.getconnection()

    def test_getbucket(self):
        conn = self._makeit(self._token)
        bucket = conn.getbucket()
        self.assertEqual(bucket.name, self._bucket)



from lacli.exceptions import ApiAuthException, BaseAppException, PauseEvent
from lacli.server.interface.ClientInterface.ttypes import InvalidOperation
from lacli.server.interface.ClientInterface.ttypes import ErrorType
from lacli.log import getLogger
from twisted.python import log as twisted_log
from twisted.python import failure
from twisted.internet import defer
from functools import wraps
import errno


def log_hide(**kwargs):
    def decorator(f):
        setattr(f, '__hidden', kwargs)
        return f
    return decorator


def log_call(f):
    @wraps(f)
    @defer.inlineCallbacks
    def w(*args, **kwargs):
        hide = getattr(f, '__hidden', {})
        if len(args) > 0 and hasattr(args[0], "debug"):
            try:
                if int(args[0].debug) > 4:
                    hide = {}  # don't hide anything
            except:
                pass
        hidden = lambda k: k in hide and hide[k] is True
        maybe = lambda k, v: "<...>" if hidden(k) else v
        _as = maybe('_args', args[1:])
        _ks = {(k, maybe(k, v)) for k, v in kwargs.iteritems()}
        getLogger().debug("calling {} with {}, {}".format(f, _as, _ks))
        r = yield f(*args, **kwargs)
        getLogger().debug("return value for {} is {}".format(f, maybe('_ret', r)))
        defer.returnValue(r)
    return w
            

def tthrow(f):
    """ Decorate a method to raise InvalidOperation instead
        of lacli exceptions
    """
    @wraps(f)
    @defer.inlineCallbacks
    def w(*args, **kwargs):
        r = None
        try:
            r = yield f(*args, **kwargs)
            defer.returnValue(r)
        except ApiAuthException as e:
            twisted_log.err(e)
            raise InvalidOperation(ErrorType.Authentication, e.msg)
        except NotImplementedError as e:
            getLogger().debug("{} is not implemented".format(f, r))
            raise InvalidOperation(ErrorType.NotImplemented, str(e))
        except OSError as e:
            getLogger().debug("{} threw exception".format(f), exc_info=True)
            if e.errno == errno.ENOENT:
                raise InvalidOperation(ErrorType.FileNotFound,
                                       "File not found",
                                       filename=e.filename)
            if e.errno == errno.EACCES:
                raise InvalidOperation(ErrorType.Other, "Access denied or file in use")
            getLogger().debug("unknown exception", exc_info=True)
            raise InvalidOperation(ErrorType.Other, "Unknown error")
        except IOError as e:
            getLogger().debug("{} threw exception".format(f), exc_info=True)
            if e.errno == errno.ENOENT:
                raise InvalidOperation(ErrorType.FileNotFound,
                                       "File not found",
                                       filename=e.filename)
            if e.errno == errno.ENOSPC:
                raise InvalidOperation(ErrorType.Other,
                                       "No space left on device")
            getLogger().debug("unknown exception", exc_info=True)
            raise InvalidOperation(ErrorType.Other, "")
        except ValueError as e:
            getLogger().debug("{} threw exception".format(f), exc_info=True)
            raise InvalidOperation(ErrorType.Validation, str(e))
        except BaseAppException as e:
            getLogger().debug("application exception", exc_info=True)
            twisted_log.err(e)
            raise InvalidOperation(ErrorType.Other, str(e))
        except PauseEvent:
            pass
        except Exception as e:
            getLogger().debug("unknown exception", exc_info=True)
            twisted_log.err(e)
            raise InvalidOperation(ErrorType.Other, "Unknown error")
    return log_call(w)



from StringIO import StringIO
import hashlib


class HashIO(StringIO):
    def __init__(self):
        StringIO.__init__(self)
        self.md5 = hashlib.md5()

    def write(self, s):
        self.md5.update(s)

    def getvalue(self):
        return self.md5.digest()



import shlex
import sys

from lacli.enc import get_unicode
from lacli.log import getLogger
from twisted.internet import defer
from twisted.python.failure import Failure
from twisted.internet import reactor
from docopt import docopt, DocoptExit
from functools import update_wrapper, wraps, partial
from requests.exceptions import ConnectionError, HTTPError
from lacli.exceptions import (ApiErrorException, ApiAuthException, CacheInitException,
                              ApiUnavailableException, ApiNoSessionError, BaseAppException)
from crochet import setup, run_in_reactor, TimeoutError

def expand_args(f):
    @wraps(f)
    def wrap(args=[], kwargs={}):
        return f(*args, **kwargs)
    return wrap


class deferred_property(object):
    def __init__(self, f):
        self.f = defer.inlineCallbacks(f)
        update_wrapper(self, self.f)

    def update(self, obj, name, value):
        if isinstance(value, Failure):
            del obj.__dict__[name]
        else:
            obj.__dict__[name]=value
        return value

    def __get__(self, obj, cls):
        if obj is None:
            return self
        name = self.f.__name__
        if name not in obj.__dict__:
            obj.__dict__[name] = self.f(obj)
            obj.__dict__[name].addBoth(
                partial(self.update, obj, name))
        return obj.__dict__[name]

    def __set__(self, obj, value):
        obj.__dict__[self.f.__name__] = value


class cached_property(object):
    val = None

    def __init__(self, f):
        self.f = f
        update_wrapper(self, self.f)

    def __get__(self, obj, cls):
        if obj is None:
            return self
        if not self.f.__name__ in obj.__dict__:
            obj.__dict__[self.f.__name__] = self.f(obj)
        return obj.__dict__[self.f.__name__]

    def __set__(self, obj, value):
        obj.__dict__[self.f.__name__] = value


def with_api_response(f):
    """ Decorate a method to capture API related errors
    """
    @wraps(f)
    def wrap(*args, **kwargs):
        try:
            r = f(*args, **kwargs)
            r.raise_for_status()
            return r.json()
        except ConnectionError as e:
            raise ApiUnavailableException(e)
        except HTTPError as e:
            if e.response.status_code == 404:
                raise ApiUnavailableException(e)
            elif e.response.status_code == 401:
                raise ApiAuthException(exc=e)
            else:
                raise ApiErrorException(e)
        except ApiNoSessionError:
            raise
        except BaseAppException:
            raise
        except Exception as e:
            raise ApiErrorException(e)
    return wrap


def contains(cls):
    def decorator(func):
        @wraps(func)
        def patched(*args, **kwargs):
            return cls(func(*args, **kwargs))
        return patched
    return decorator


def command(**types):
    """ Decorator to parse command options with docopt and
        validate the types.
    """
    def decorate(func):
        @wraps(func)
        def wrap(self, line):
            kwargs = {}
            try:
                opts = docopt(func.__doc__, shlex.split(line))
                for opt, val in opts.iteritems():
                    kw = opt.strip('<>')
                    if val and kw in types:
                        if types[kw] == unicode:
                            kwargs[kw] = get_unicode(val)
                        else:
                            kwargs[kw] = types[kw](val)
            except ValueError as e:
                print "error: invalid value:", e
                print func.__doc__
                return
            except DocoptExit as e:
                print e
                return
            try:
                func(self, **kwargs)
            except CacheInitException as e:
                getLogger().debug("Cache not initialized", exc_info=True)
                print "Could not initialize cache"
                return
        return wrap
    return decorate


def block(f):
    """ Decorate a method to block in crochet reactor
    """
    fblocking = run_in_reactor(f)
    @wraps(f)
    def wrap(*args, **kwargs):
        if not reactor.running:
            setup()
        result = fblocking(*args, **kwargs)
        try:
            return result.wait()
        except TimeoutError:
            result.cancel()
            raise
    return wrap


class login_async(object):
    def __init__(self, f, obj=None):
        self.f = f
        self.obj = obj
        update_wrapper(self, self.f)

    def __get__(self, obj, cls):
        return wraps(self.f)(login(self.f, obj))

    def dologin(self, prefs):
        return self.obj.registry.cmd.login.login_async(
            prefs.get('user'), prefs.get('pass'))

    @defer.inlineCallbacks
    def loginfirst(self, prefs, *args, **kwargs):
        yield self.dologin(prefs)
        r = yield self.f(self.obj, *args, **kwargs)
        defer.returnValue(r)

    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            if self.obj is None:
                self.obj = args[0]
                args = args[1:]
            elif self.obj == args[0]:
                args = args[1:]

        prefs = None
        if self.obj.registry.session:
            prefs = self.obj.registry.session.prefs
        if not prefs:
            prefs = self.obj.registry.init_prefs()

        if not self.obj.session or self.obj.registry.cmd.login.email is None:
            return self.loginfirst(prefs, *args, **kwargs)
        return self.f(self.obj, *args, **kwargs)


class login(login_async):
    def __init__(self, *args, **kwargs):
        super(login, self).__init__(*args, **kwargs)
    
    def dologin(self, prefs):
        if self.obj.batch:
            self.obj.registry.cmd.login.login_batch(
                prefs.get('user'), prefs.get('pass'))
        else:
            cmdline = [prefs[a] for a in ['user', 'pass']
                       if prefs.get(a)]
            self.obj.registry.cmd.username = prefs['user']
            self.obj.registry.cmd.password = prefs['pass']
            self.obj.registry.cmd.do_login(" ".join(cmdline))

    def loginfirst(self, *args, **kwargs):
       return super(login, self).loginfirst(*args, **kwargs)
# vim: et:sw=4:ts=4



from testtools import TestCase
from mock import patch


class ExceptionsTest(TestCase):
    def setUp(self):
        super(ExceptionsTest, self).setUp()

    def tearDown(self):
        super(ExceptionsTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacli.exceptions import BaseAppException
        return BaseAppException(*args, **kw)

    def test_exception(self):
        assert self._makeit(Exception())

    def test_str(self):
        exc = self._makeit(Exception("lala"))
        exc.msg = "foobar"
        self.assertEqual("foobar", str(exc))

    def test_worker_failure(self):
        import lacli.exceptions
        with patch.object(lacli.exceptions, 'current_process',
                          create=True) as p:
            p.return_value = "foo"
            exc = lacli.exceptions.WorkerFailureError(Exception("lala"))
            self.assertEqual(p.call_count, 1)
            self.assertEqual("worker 'foo' failed", str(exc))



import os
from sys import maxint, stderr
from progressbar import (ProgressBar, Bar,
                         ETA, FileTransferSpeed)
from lacli.log import queueHandler, getLogger
from abc import ABCMeta, abstractmethod


class BaseProgressHandler(queueHandler, ProgressBar):
    __metaclass__ = ABCMeta
    
    def __init__(self, *args, **kwargs):
        self.tx = {}
        self.progress = kwargs.pop('progress', 0)
        super(BaseProgressHandler, self).__init__(*args, **kwargs)

    def update_current(self, msg):
        self.tx[msg['part']] = int(msg['tx'])
        return sum(self.tx.values())

    def save_current(self):
        self.progress += sum(self.tx.values())
        self.tx = {}

    def handle(self, msg):
        if 'part' in msg:
            self.update(self.progress + self.update_current(msg))
        elif 'save' in msg:
            self.keydone(msg)
            self.save_current()

    def __enter__(self):
        q = super(BaseProgressHandler, self).__enter__()
        progressToQueue(q)
        self.start(initval=self.progress)
        return q

    @abstractmethod
    def keydone(self, msg):
        getLogger().debug("saved key: " + str(msg['key'])
            + " (" + str(msg['size']) + ")")


class StateProgressHandler(BaseProgressHandler):
    uploads = {}

    def __init__(self, state=None, **kwargs):
        assert state is not None, "StateProgressHandler requires a state object"
        self.state = state
        progress = kwargs.pop('progress', 0)
        for seq, key in enumerate(self.state.keys):
            progress += key['size']
        kwargs['progress'] = progress
        super(StateProgressHandler, self).__init__(**kwargs)
        
    def update(self, value=None):
        if value is not None:
            self.state.update(value)
        super(StateProgressHandler, self).update(value)
        getLogger().debug("State progress: {}".format(self.state.progress))

    def __enter__(self):
        type(self).uploads[self.state] = self
        return super(StateProgressHandler, self).__enter__()

    def __exit__(self, cls, value, traceback):
        del type(self).uploads[self.state]

    def keydone(self, msg):
        getLogger().debug("saving key " + str(msg['key']) + " to state file")
        self.state.keydone(msg['key'], msg['size'])


class ConsoleProgressHandler(StateProgressHandler):
    def __init__(self, *args, **kwargs):
        fname = kwargs.pop('fname', "-")
        kwargs.setdefault('widgets', [ fname, ' : ', Bar(), ' ', ETA(), ' ', FileTransferSpeed()])
        super(ConsoleProgressHandler, self).__init__(*args, **kwargs)

    def __enter__(self):
        self.start()
        return super(ConsoleProgressHandler, self).__enter__()

    def update(self, value=None):
        super(ConsoleProgressHandler, self).update(value)
        self.fd.flush()


class ServerProgressHandler(StateProgressHandler):
    def __init__(self, *args, **kwargs):
        self.eta = ""
        kwargs.setdefault('widgets', [])
        kwargs.setdefault('term_width', 0)
        if 'fd' not in kwargs:
            kwargs['fd'] = open(os.devnull, 'w+')
        super(ServerProgressHandler, self).__init__(*args, **kwargs)

    def eta(self):
        if self.currval == 0:
            return ''
        else:
            elapsed = self.seconds_elapsed
            eta = elapsed * self.maxval / self.currval - elapsed
            return self.format_time(eta)

progress = None


def progressToQueue(queue):
    global progress
    progress = queue


def make_progress(msg):
    global progress
    progress.put(msg)


def save_progress(key, size):
    global progress
    progress.put({'save': True, 'key': key, 'size': size})



import os

from testtools import TestCase
from testtools.matchers import Contains
from . import makeprefs
from mock import Mock, patch
from StringIO import StringIO


class LoginCommandTest(TestCase):
    def setUp(self):
        super(LoginCommandTest, self).setUp()
        self.prefs = makeprefs()
        self.home = os.path.join('t', 'data', 'home')

    def tearDown(self):
        super(LoginCommandTest, self).tearDown()

    def _makeit(self, *args, **kwargs):
        from lacli.main import LaLoginCommand
        return LaLoginCommand(*args, **kwargs)

    def test_login_command(self):
        assert self._makeit(Mock(), Mock(), self.prefs)

    @patch('sys.stdout', new_callable=StringIO)
    def test_do_login(self, stdout):
        registry = Mock()
        registry.prefs = self.prefs
        registry.new_session.return_value.account = {'email': 'foo'}
        cli = self._makeit(registry)

        cli.onecmd('login')
        self.assertThat(stdout.getvalue(),
                        Contains('authentication succesfull'))



import os
from lacli.log import getLogger
from lacli.exceptions import PauseEvent
from multiprocessing import Queue, active_children
from Queue import Empty


class ControlHandler(object):

    def __init__(self):
        self.q = None

    def __enter__(self):
        self.q = Queue()
        controlByQueue(self.q)
        return self.q

    def __exit__(self, type, value, traceback):
        getLogger().debug("control handler exiting.")
        if type == KeyboardInterrupt:
            getLogger().debug("upload pausing.")
            self.pause()
            type = None
        stopControlByQueue()
        return type is None

    def pause(self):
        if self.q is not None:
            children = len(active_children())
            getLogger().debug(
               "ControlHandler pausing {} children".format(children))
            map(self.q.put, [{'pause': True}] * children)
        else:
            getLogger().debug(
               "ControlHandler.pause() called " +
               "when no control context in effect")
                

controlq = None


def controlByQueue(queue):
    global controlq
    controlq = queue


def stopControlByQueue():
    global controlq
    controlq = None

def readControl():
    global controlq
    try:
        msg = controlq.get(False)
        if 'pause' in msg:
            raise PauseEvent()
    except Empty:
        pass
# vim: et:sw=4:ts=4



# -*- coding: utf-8 -*-

# Resource object code
#
# Created: Mon Nov 11 13:59:39 2013
#      by: The Resource Compiler for PySide (Qt v4.8.5)
#
# WARNING! All changes made in this file will be lost!

from PySide import QtCore

qt_resource_data = "\x00\x00\x06\xdcimport QtQuick 1.1\x0d\x0aFocusScope {\x0d\x0a    id: focusScope\x0d\x0a    property int fontSize: 12\x0d\x0a    property int textBoxWidth: parent.width * 0.8\x0d\x0a    property int textBoxHeight: 45\x0d\x0a    property alias text: textInput.text\x0d\x0a    property alias acceptableInput: textInput.acceptableInput\x0d\x0a    property alias font: textInput.font\x0d\x0a    property alias inputMask: textInput.inputMask\x0d\x0a    signal textChanged(string text)\x0d\x0a    width: textBoxWidth\x0d\x0a    height: textBoxHeight\x0d\x0a\x0d\x0a    Rectangle {\x0d\x0a        width: parent.width\x0d\x0a        height: parent.height\x0d\x0a        color: \x22#d7d7d7\x22\x0d\x0a        border.color:\x22#000000\x22\x0d\x0a        border.width: 3\x0d\x0a        radius: 0\x0d\x0a        MouseArea {\x0d\x0a            anchors.fill: parent\x0d\x0a            onClicked: {\x0d\x0a                focusScope.focus = true\x0d\x0a                textInput.openSoftwareInputPanel()\x0d\x0a            }\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a    MouseArea {\x0d\x0a        anchors.fill: parent\x0d\x0a        onClicked: {\x0d\x0a            focusScope.focus = true\x0d\x0a            textInput.openSoftwareInputPanel()\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a    TextInput {\x0d\x0a        id: textInput\x0d\x0a        anchors {\x0d\x0a            right: parent.right\x0d\x0a            rightMargin: 8\x0d\x0a            verticalCenter: parent.verticalCenter\x0d\x0a        }\x0d\x0a        focus: true\x0d\x0a        selectByMouse: true\x0d\x0a        font.pointSize: fontSize\x0d\x0a        onTextChanged: focusScope.textChanged(text)\x0d\x0a        anchors.left: parent.left\x0d\x0a        anchors.leftMargin: 8\x0d\x0a    }\x0d\x0a\x0d\x0a    transitions: [\x0d\x0a        Transition {\x0d\x0a            from: ''; to: 'hasText'\x0d\x0a            NumberAnimation {\x0d\x0a                properties: 'opacity'\x0d\x0a            }\x0d\x0a        },\x0d\x0a        Transition {\x0d\x0a            from: 'hasText'; to: ''\x0d\x0a            NumberAnimation {\x0d\x0a                properties: 'opacity'\x0d\x0a            }\x0d\x0a        }\x0d\x0a    ]\x0d\x0a}\x0d\x0a\x00\x00\x05\xe5import QtQuick 1.1\x0d\x0aFocusScope {\x0d\x0a    width: row1.width\x0d\x0a    height: row1.height\x0d\x0a    Grid {\x0d\x0a        id: row1\x0d\x0a        columns: 2\x0d\x0a        rows: 1\x0d\x0a        spacing: 10\x0d\x0a        Rectangle {\x0d\x0a            x: 0\x0d\x0a            y: 0\x0d\x0a            width: text1.width + 10\x0d\x0a            height: text_input1.height+5\x0d\x0a            color: \x22#000000\x22\x0d\x0a            radius: 99\x0d\x0a            border.width: 1\x0d\x0a            border.color: \x22#000000\x22\x0d\x0a\x0d\x0a            Text {\x0d\x0a                id: text1\x0d\x0a                x: 8\x0d\x0a                y: 0\x0d\x0a                color: \x22#ffffff\x22\x0d\x0a\x0d\x0a                text: \x22A\x22\x0d\x0a                anchors.verticalCenterOffset: 0\x0d\x0a                anchors.horizontalCenterOffset: 0\x0d\x0a                anchors.topMargin: 0\x0d\x0a                anchors.horizontalCenter: parent.horizontalCenter\x0d\x0a                anchors.verticalCenter: parent.verticalCenter\x0d\x0a                anchors.top: parent.top\x0d\x0a                font.pointSize: 11\x0d\x0a                verticalAlignment: Text.AlignVCenter\x0d\x0a            }\x0d\x0a        }\x0d\x0a        Rectangle {\x0d\x0a            width: text_input1.width +5\x0d\x0a            height: text_input1.height+5\x0d\x0a            color: \x22#00000000\x22\x0d\x0a\x0d\x0a            Text {\x0d\x0a                id: text_input1\x0d\x0a                focus: true\x0d\x0a                anchors.horizontalCenter: parent.horizontalCenter\x0d\x0a                anchors.verticalCenter: parent.verticalCenter\x0d\x0a                font.pointSize: 12\x0d\x0a                font.family: \x22Consolas\x22\x0d\x0a                text: \x22your key id\x22\x0d\x0a\x0d\x0a            }\x0d\x0a        }\x0d\x0a    }\x0d\x0a}\x0d\x0a\x00\x00\x06\xe8import QtQuick 1.1\x0d\x0a\x0d\x0aRectangle {\x0d\x0a    id: button\x0d\x0a\x0d\x0a    property alias text: textItem.text\x0d\x0a    property bool enabled: false\x0d\x0a    property alias radius: button.radius\x0d\x0a    property alias font: textItem.font\x0d\x0a    signal buttonClicked()\x0d\x0a\x0d\x0a    width: textItem.width + 40; height: textItem.height + 10\x0d\x0a    border.width: 1\x0d\x0a    radius: height/4\x0d\x0a    smooth: true\x0d\x0a\x0d\x0a    gradient: Gradient {\x0d\x0a        GradientStop { id: topGrad; position: 0.0; color: \x22#333\x22 }\x0d\x0a        GradientStop { id: bottomGrad; position: 1.0; color: \x22#000\x22 }\x0d\x0a    }\x0d\x0a\x0d\x0a    Text {\x0d\x0a        id: textItem\x0d\x0a        x: parent.width/2 - width/2; y: parent.height/2 - height/2\x0d\x0a        font.pixelSize: 18\x0d\x0a        color: \x22black\x22\x0d\x0a        font.family: \x22Open Sans\x22\x0d\x0a        style: Text.Raised\x0d\x0a    }\x0d\x0a\x0d\x0a    MouseArea {\x0d\x0a        id: mouseArea\x0d\x0a        anchors.fill: parent\x0d\x0a        enabled: false\x0d\x0a    }\x0d\x0a\x0d\x0a    states: [\x0d\x0a    State {\x0d\x0a        name: \x22pressed\x22; when: mouseArea.pressed && mouseArea.containsMouse\x0d\x0a        PropertyChanges { target: topGrad; color: \x22white\x22 }\x0d\x0a        PropertyChanges { target: bottomGrad; color: \x22white\x22 }\x0d\x0a        PropertyChanges { target: textItem; x: textItem.x + 1; y: textItem.y + 1; font.family: \x22Open Sans\x22; explicit: true }\x0d\x0a        PropertyChanges { target: mouseArea; enabled: true; onClicked: button.buttonClicked() }\x0d\x0a    },\x0d\x0a    State {\x0d\x0a        name: \x22Enabled\x22\x0d\x0a        when: button.enabled\x0d\x0a        PropertyChanges { target: topGrad; position: 0; color: \x22white\x22 }\x0d\x0a        PropertyChanges { target: bottomGrad; position: 1; color: \x22white\x22 }\x0d\x0a        PropertyChanges { target: mouseArea; enabled: true; onClicked: button.buttonClicked() }\x0d\x0a\x0d\x0a        PropertyChanges {\x0d\x0a            target: textItem\x0d\x0a            font.family: \x22Open Sans\x22\x0d\x0a        }\x0d\x0a    }\x0d\x0a    ]\x0d\x0a}\x0d\x0a\x00\x00\x17Bimport QtQuick 1.1\x0d\x0a\x0d\x0a\x0d\x0aColumn {\x0d\x0a    id: keyinput\x0d\x0a    signal decrypt(string key, string path)\x0d\x0a    property bool hasError: false\x0d\x0a    transformOrigin: Item.Center\x0d\x0a\x0d\x0a    spacing: 20\x0d\x0a\x0d\x0a    ListModel {\x0d\x0a        id: mymodel\x0d\x0a        ListElement {\x0d\x0a            name: \x22B\x22\x0d\x0a            complete: false\x0d\x0a            value: \x22\x22\x0d\x0a        }\x0d\x0a        ListElement {\x0d\x0a            name: \x22C\x22\x0d\x0a            complete: false\x0d\x0a            value: \x22\x22\x0d\x0a        }\x0d\x0a        ListElement {\x0d\x0a            name: \x22D\x22\x0d\x0a            complete: false\x0d\x0a            value:\x22\x22\x0d\x0a        }\x0d\x0a        ListElement {\x0d\x0a            name: \x22E\x22\x0d\x0a            complete: false\x0d\x0a            value:\x22\x22\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a    Component {\x0d\x0a        id: _error_msg\x0d\x0a        Column {\x0d\x0a            Text {\x0d\x0a                color: \x22#ff0000\x22\x0d\x0a                text: \x22Error\x22\x0d\x0a                anchors.right: parent.right\x0d\x0a                anchors.rightMargin: 0\x0d\x0a                font.pointSize: 15\x0d\x0a            }\x0d\x0a            Text {\x0d\x0a                color: \x22#992626\x22\x0d\x0a                text: \x22please try again\x22\x0d\x0a                font.pointSize: 9\x0d\x0a                font.family: \x22Open Sans\x22\x0d\x0a                verticalAlignment: Text.AlignVCenter\x0d\x0a            }\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a    Row {\x0d\x0a        height: 1\x0d\x0a        anchors.right: parent.right\x0d\x0a        anchors.rightMargin: 0\x0d\x0a        anchors.left: parent.left\x0d\x0a        anchors.leftMargin: 0\x0d\x0a    }\x0d\x0a\x0d\x0a    Row {\x0d\x0a        anchors.horizontalCenter: parent.horizontalCenter\x0d\x0a        Text {\x0d\x0a            id: list_title\x0d\x0a            text: \x22Enter Certificate Key\x22\x0d\x0a            font.family: \x22Arial\x22\x0d\x0a            font.bold: true\x0d\x0a            font.pixelSize: 19\x0d\x0a            verticalAlignment: Text.AlignVCenter\x0d\x0a            horizontalAlignment: Text.AlignHCenter\x0d\x0a        }\x0d\x0a    }\x0d\x0a    Row {\x0d\x0a        id: key_row\x0d\x0a        Column {\x0d\x0a            id: column1\x0d\x0a            spacing: 10\x0d\x0a\x0d\x0a            Repeater {\x0d\x0a                id: list\x0d\x0a                focus: true\x0d\x0a                model: mymodel\x0d\x0a                onItemAdded: {\x0d\x0a                    if (!index) {\x0d\x0a                        item.forceActiveFocus()\x0d\x0a                    }\x0d\x0a                }\x0d\x0a\x0d\x0a                delegate: KeyRow {\x0d\x0a                    id: foo\x0d\x0a                    Connections {\x0d\x0a                        onAcceptableChanged: {\x0d\x0a                            mymodel.setProperty(index, 'complete', foo.acceptable)\x0d\x0a                            button1.rowComplete()\x0d\x0a                            if (foo.acceptable)\x0d\x0a                                if (index + 1 < list.count)\x0d\x0a                                    list.itemAt(index+1).forceActiveFocus()\x0d\x0a                                else\x0d\x0a                                    folder.forceActiveFocus()\x0d\x0a                                keyinput.hasError = false\x0d\x0a                        }\x0d\x0a\x0d\x0a                        onRowChanged: {\x0d\x0a                            mymodel.setProperty(index, 'value', val)\x0d\x0a                        }\x0d\x0a                    }\x0d\x0a                }\x0d\x0a            }\x0d\x0a        }\x0d\x0a    }\x0d\x0a    Row{\x0d\x0a        id: folder_row\x0d\x0a        anchors.right: parent.right\x0d\x0a        anchors.rightMargin: 10\x0d\x0a        layoutDirection: Qt.RightToLeft\x0d\x0a        spacing: 10\x0d\x0a\x0d\x0a\x0d\x0a        Button {\x0d\x0a            id: button2\x0d\x0a            text: \x22Browse\x22\x0d\x0a            enabled: true\x0d\x0a            Connections {\x0d\x0a                onButtonClicked: folder.text = destsel.getDirectory()\x0d\x0a            }\x0d\x0a            font.pixelSize: 10\x0d\x0a            radius: 0\x0d\x0a        }\x0d\x0a\x0d\x0a        TextBox {\x0d\x0a            id: folder\x0d\x0a            anchors.verticalCenter: parent.verticalCenter\x0d\x0a            fontSize: 12\x0d\x0a            textBoxHeight: 20\x0d\x0a            textBoxWidth: 150\x0d\x0a            focus: true\x0d\x0a            font.pointSize: 12\x0d\x0a            Connections {\x0d\x0a                onTextChanged: button1.folderChanged()\x0d\x0a            }\x0d\x0a\x0d\x0a        }\x0d\x0a\x0d\x0a        Rectangle {\x0d\x0a            id: rectangle1\x0d\x0a            width: childrenRect.width\x0d\x0a            height: childrenRect.height\x0d\x0a            color: \x22#ffffff\x22\x0d\x0a            anchors.verticalCenter: parent.verticalCenter\x0d\x0a            z: -2\x0d\x0a\x0d\x0a            Text {\x0d\x0a                id: text1\x0d\x0a                text: qsTr(\x22Folder:\x22)\x0d\x0a                font.pointSize: 12\x0d\x0a            }\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a\x0d\x0a    Row {\x0d\x0a        id: button_row\x0d\x0a        layoutDirection: Qt.RightToLeft\x0d\x0a        anchors.right: parent.right\x0d\x0a        anchors.rightMargin: 10\x0d\x0a        anchors.left: parent.left\x0d\x0a        anchors.leftMargin: 10\x0d\x0a        spacing: 10\x0d\x0a        Button {\x0d\x0a            id: button1\x0d\x0a            text: \x22Decrypt\x22\x0d\x0a            signal rowComplete()\x0d\x0a            signal folderChanged()\x0d\x0a            onRowComplete: {\x0d\x0a                enabled = destsel.dirExists(folder.text)\x0d\x0a                for (var i=0; i< mymodel.count; i++) {\x0d\x0a                    if (!mymodel.get(i).complete)\x0d\x0a                        enabled = false\x0d\x0a                }\x0d\x0a            }\x0d\x0a            onFolderChanged: {\x0d\x0a                if (destsel.dirExists(folder.text))\x0d\x0a                    rowComplete()\x0d\x0a            }\x0d\x0a\x0d\x0a            onButtonClicked: {\x0d\x0a                var key=''\x0d\x0a                for (var i=0; i<list.model.count; i++)\x0d\x0a                    key += list.model.get(i).value\x0d\x0a                keyinput.decrypt(key, folder.text)\x0d\x0a                if(!keyinput.hasError)\x0d\x0a                    Qt.quit()\x0d\x0a            }\x0d\x0a        }\x0d\x0a        Button {\x0d\x0a            id: button3\x0d\x0a            text: \x22Cancel\x22\x0d\x0a            enabled: true\x0d\x0a            onButtonClicked: Qt.quit()\x0d\x0a        }\x0d\x0a        Loader {\x0d\x0a            sourceComponent: _error_msg\x0d\x0a            id: error\x0d\x0a            opacity: 0\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a    Row {\x0d\x0a        height: 1\x0d\x0a        anchors.right: parent.right\x0d\x0a        anchors.rightMargin: 0\x0d\x0a        anchors.left: parent.left\x0d\x0a        anchors.leftMargin: 0\x0d\x0a    }\x0d\x0a    states: [\x0d\x0a        State {\x0d\x0a            when: keyinput.hasError\x0d\x0a            PropertyChanges {\x0d\x0a                target: error\x0d\x0a                opacity: 1\x0d\x0a            }\x0d\x0a        }\x0d\x0a    ]\x0d\x0a\x0d\x0a}\x0d\x0a\x0d\x0a\x00\x00\x08\x92import QtQuick 1.1\x0d\x0a\x0d\x0a\x0d\x0aFocusScope {\x0d\x0a    id: rect\x0d\x0a    property bool complete: false\x0d\x0a    property bool acceptable: textbox.acceptableInput\x0d\x0a    signal rowChanged(string val)\x0d\x0a    width: childrenRect.width\x0d\x0a    height: childrenRect.height\x0d\x0a\x0d\x0a    Row {\x0d\x0a        id: row1\x0d\x0a        layoutDirection: Qt.RightToLeft\x0d\x0a        spacing: 0\x0d\x0a\x0d\x0a        Rectangle {\x0d\x0a            width: 10\x0d\x0a            height: textbox.height\x0d\x0a            color: \x22#ffffff\x22\x0d\x0a        }\x0d\x0a        TextBox {\x0d\x0a            id: textbox\x0d\x0a            fontSize: text_proto.font.pointSize\x0d\x0a            textBoxHeight: text_proto.height\x0d\x0a            textBoxWidth: text_proto.width + 16\x0d\x0a            focus: true\x0d\x0a            font: text_proto.font\x0d\x0a            inputMask: \x22HH HH HH HH . HH HH HH HH\x22\x0d\x0a            Connections {\x0d\x0a                onTextChanged: rect.rowChanged(text.replace(/[ .]/g, ''))\x0d\x0a            }\x0d\x0a        }\x0d\x0a\x0d\x0a        Rectangle {\x0d\x0a            width: 10\x0d\x0a            height: textbox.height\x0d\x0a            color: \x22#ffffff\x22\x0d\x0a        }\x0d\x0a        Rectangle {\x0d\x0a            width: 30\x0d\x0a            color: \x22#ffffff\x22\x0d\x0a            anchors.bottom: parent.bottom\x0d\x0a            anchors.bottomMargin: 0\x0d\x0a            anchors.top: parent.top\x0d\x0a            anchors.topMargin: 0\x0d\x0a\x0d\x0a            Rectangle {\x0d\x0a                width: text1.width + 10\x0d\x0a                height: 20\x0d\x0a                color: \x22#000000\x22\x0d\x0a                anchors.verticalCenter: parent.verticalCenter\x0d\x0a                anchors.horizontalCenter: parent.horizontalCenter\x0d\x0a\x0d\x0a                Text {\x0d\x0a                    id: text1\x0d\x0a                    color: \x22#ffffff\x22\x0d\x0a                    text: name\x0d\x0a                    anchors.horizontalCenter: parent.horizontalCenter\x0d\x0a                    font.pixelSize: 15\x0d\x0a                    transformOrigin: Item.Center\x0d\x0a                    anchors.verticalCenter: parent.verticalCenter\x0d\x0a                    verticalAlignment: Text.AlignVCenter\x0d\x0a                }\x0d\x0a            }\x0d\x0a        }\x0d\x0a    }\x0d\x0a\x0d\x0a    TextInput {\x0d\x0a        id: text_proto\x0d\x0a        visible: false\x0d\x0a        font.pointSize: 15\x0d\x0a        font.family: \x22Consolas\x22\x0d\x0a        font.capitalization: Font.AllUppercase\x0d\x0a        text: qsTr(\x22ff ff ff ff . ff ff ff ff\x22)\x0d\x0a\x0d\x0a    }\x0d\x0a}\x0d\x0a"
qt_resource_name = "\x00\x02\x00\x00\x07\xb9\x00u\x00i\x00\x0b\x09\xde\xcb<\x00T\x00e\x00x\x00t\x00B\x00o\x00x\x00.\x00q\x00m\x00l\x00\x09\x09j\xcb\x1c\x00I\x00d\x00R\x00o\x00w\x00.\x00q\x00m\x00l\x00\x0a\x0bhq\x5c\x00B\x00u\x00t\x00t\x00o\x00n\x00.\x00q\x00m\x00l\x00\x0c\x0d\xfe\x98\x9c\x00K\x00e\x00y\x00I\x00n\x00p\x00u\x00t\x00.\x00q\x00m\x00l\x00\x0a\x09dq\xfc\x00K\x00e\x00y\x00R\x00o\x00w\x00.\x00q\x00m\x00l"
qt_resource_struct = "\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x02\x00\x00\x00\x05\x00\x00\x00\x02\x00\x00\x00v\x00\x00\x00\x00\x00\x01\x00\x00*\xfb\x00\x00\x00&\x00\x00\x00\x00\x00\x01\x00\x00\x06\xe0\x00\x00\x00\x0a\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00>\x00\x00\x00\x00\x00\x01\x00\x00\x0c\xc9\x00\x00\x00X\x00\x00\x00\x00\x00\x01\x00\x00\x13\xb5"
def qInitResources():
    QtCore.qRegisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



import json
import collections
import errno
import os
import shlex
import time
from urlparse import urlparse
from pkg_resources import resource_string
from lacli.date import parse_timestamp, later

from glob import iglob
from lacli.adf import (load_archive, make_adf, Certificate, Archive,
                       Meta, Links, Cipher, Signature, as_json)
from lacli.log import getLogger
from lacli.archive import dump_archive, archive_handle
from lacli.exceptions import InvalidArchiveError, CacheInitException
from lacli.decorators import contains
from lacli.server.interface.ClientInterface.ttypes import ArchiveStatus
from urllib import pathname2url
from tempfile import NamedTemporaryFile
from binascii import b2a_hex
from itertools import izip, imap
from subprocess import check_call
from contextlib import contextmanager


def group(it, n, dl):
    return imap(dl.join, izip(*[it]*n))


def pairs(it, dl=""):
    return group(it, 2, dl)


def fours(it, dl=" "):
    return group(it, 4, dl)


class Cache(object):
    def __init__(self, home):
        self.home = home

    def _cache_dir(self, path, write=False):
        if self.home is None:
            raise CacheInitException()
        dname = os.path.join(self.home, path)
        if not os.path.exists(dname) and write:
            os.makedirs(dname, mode=0744)
        return dname

    def _archive_open(self, name, mode='r'):
        dname = self._cache_dir('archives', write='w' in mode)
        return open(os.path.join(dname, name), mode)

    def _cert_open(self, name, mode='r'):
        dname = self._cache_dir('certs', write='w' in mode)
        return open(os.path.join(dname, name), mode)

    def _upload_open(self, name, mode='r'):
        dname = self._cache_dir('uploads', write='w' in mode)
        return open(os.path.join(dname, name), mode)

    def get_adf(self, fname, category='archives'):
        with open(os.path.join(self._cache_dir(category), fname)) as f:
            return load_archive(f)

    def _validate_upload(self, lines):
        parts = []
        ret = {}
        last = chunk = False
        for line in lines:
            key = json.loads(line)
            if 'uri' in key:
                ret.update(key)
                continue
            size = key['size']
            if chunk is False:  # initialize chunk size
                chunk = last = size
            if size < last and last == chunk:
                last = size  # chunk is smaller than all previous
            assert size == chunk or size == last, "Too many different sizes"
            parts.append(key)
        ret['keys'] = parts
        return ret

    def _get_uploads(self):
        uploads = {}
        for fn in iglob(os.path.join(self._cache_dir('uploads'), '*')):
            aid = os.path.basename(fn)
            uploads[aid] = {}
            with open(fn) as f:
                try:
                    uploads[aid] = self._validate_upload(f)
                except Exception as e:
                    getLogger().debug("Error validating upload {}".format(fn), exc_info=True)
        return uploads
                     
    def _del_upload(self, archive):
        os.unlink(os.path.join(self._cache_dir('uploads'), archive))

    def _write_upload(self, uri, capsule, logfile, exc=None, paused=False):
        new = { 'uri': uri, 'exc': exc, 'paused': paused }
        if capsule is not None:
            ks = ('resource_uri', 'size', 'title', 'remaining', 'id')
            new['capsule'] = {k: capsule.get(k, None) for k in ks}
        logfile.write(json.dumps(new)+"\n")
        logfile.flush()

    def _checkpoint_upload(self, key, size, logfile):
        new = { 'name': key, 'size': size}
        logfile.write(json.dumps(new)+"\n")
        logfile.flush()
        return new

    @contains(dict)
    def _for_adf(self, category):
        for fn in iglob(os.path.join(self._cache_dir(category), '*.adf')):
            with open(fn) as f:
                try:
                    yield (os.path.basename(fn), load_archive(f))
                except InvalidArchiveError:
                    getLogger().debug(fn, exc_info=True)

    def prepare(self, title, items, description=None, fmt='zip', cb=None):
        docs = {
            'archive': Archive(title, Meta(fmt, Cipher('aes-256-ctr', 1)),
                               description=description),
            'cert': Certificate(),
        }
        tmpdir = self._cache_dir('data', write=True)
        if isinstance(items, basestring):
            items = [items]

        def mycb(path, rel):
            if cb is not None:
                cb(path, rel)
            elif not path:
                print "Encrypting.."
            else:
                print path.encode('utf8')
        name, path, docs['auth'] = dump_archive(
            docs['archive'], items, docs['cert'], mycb, tmpdir)
        rel = os.path.relpath(path, self.home)
        docs['links'] = Links(local=pathname2url(rel))
        docs['archive'].meta.size = os.path.getsize(path)
        tmpargs = {'delete': False,
                   'dir': self._cache_dir('archives', write=True)}
        with NamedTemporaryFile(prefix=name, suffix=".adf", **tmpargs) as f:
            make_adf(list(docs.itervalues()), out=f)
            return (f.name, docs)

    def archive_status(self, fname, docs):
        if 'signature' in docs:
            return ArchiveStatus.Completed
        else:
            uploads = self._get_uploads()
            if fname in uploads:
                upload = uploads[fname]
                if "exc" in upload and upload["exc"] is not None:
                    return ArchiveStatus.Failed
                if "paused" in upload and upload["paused"] is True:
                    return ArchiveStatus.Paused
                return ArchiveStatus.Local
            elif 'links' in docs and docs['links'].upload:
                return ArchiveStatus.Paused
        return ArchiveStatus.Local

    def save_upload(self, fname, docs, uri=None, account=None):
        if uri is not None:
            docs['links'].upload = uri 
            if account is not None:
                docs['archive'].meta.email = account['email']
                docs['archive'].meta.name = account['displayname']
            with self._archive_open(fname, 'w') as f:
                make_adf(list(docs.itervalues()), out=f)
        return {
            'fname': fname,
            'link': docs['links'].upload,
            'archive': docs['archive']
        }

    def upload_error(self, fname):
        docs = []
        with self._archive_open(fname) as _upload:
            docs = load_archive(_upload)
        docs['links'].upload = None
        with self._archive_open(fname, 'w') as _upload:
            make_adf(list(docs.itervalues()), out=_upload)
        return docs

    def upload_complete(self, fname, status):
        """
        adds the archive key to the signature part of the ADF file
        """
        assert 'archive_key' in status, "no archive key"
        docs = []
        with self._archive_open(fname) as _upload:
            docs = load_archive(_upload)
        docs['signature'] = Signature(aid=status['archive_key'],
                                      uri=status.get('archive'),
                                      expires=status.get('expires'),
                                      created=status.get('created'),
                                      creator=docs['archive'].meta.name)
        with self._archive_open(fname, 'w') as _upload:
            make_adf(list(docs.itervalues()), out=_upload)
        return docs

    def save_cert(self, docs):
        """
        write cert documents to the cache directory under a unique filename.
        """
        fname = archive_handle(list(docs.itervalues()))
        tmpargs = {'delete': False,
                   'dir': self._cache_dir('certs', write=True),
                   'suffix': ".adf",
                   'prefix': fname}
        certs = self.certs()
        aid = docs['signature'].aid  # let's check if we already have this cert
        uri = docs['signature'].uri
        if not aid in certs or uri != certs[aid]['signature'].uri:
            with NamedTemporaryFile(**tmpargs) as f:
                make_adf(list(docs.itervalues()), out=f)
                return (aid, f.name)
        else:
            return (aid, None)

    def import_cert(self, fname):
        with open(fname) as cert:
            docs = load_archive(cert)
            return self.save_cert(docs)

    def _printable_cert(self, docs):
        archive = docs['archive']
        cipher = archive.meta.cipher
        if hasattr(cipher, 'mode'):
            cipher = cipher.mode
        created = parse_timestamp(archive.meta.created)
        expires = None
        aid = "-"
        if 'signature' in docs:
            aid = docs['signature'].aid or aid
            if docs['signature'].expires is not None:
                expires = parse_timestamp(docs['signature'].expires)
            if docs['signature'].created is not None:
                created = parse_timestamp(docs['signature'].created)
        md5 = b2a_hex(docs['auth'].md5).upper()
        key = b2a_hex(docs['cert'].key).upper()
        hk = pairs(fours(pairs(iter(key))), " . ")

        return unicode(resource_string(__name__, "data/certificate.html")).format(
            json=as_json(docs),
            aid=aid,
            keyB=next(hk),
            keyC=next(hk),
            keyD=next(hk),
            keyE=next(hk),
            name=archive.meta.name or "",
            email=archive.meta.email or "",
            uploaded=created.strftime("%b %d, %Y"),
            expires=(expires and expires.strftime("%b %d, %Y")) or "unknown",
            title=archive.title or "",
	    size=archive.meta.size,
            desc=archive.description or "",
	    md5=md5,
	    fmt=archive.meta.format,
            cipher=cipher).encode('utf8')

    def shred_file(self, fname, srm=None, insecure=False):
        commands = []
        if srm:
            commands.append(srm)
        else:
            commands.append('srm.bat')
            commands.append('srm')
            commands.append('shred -u')
            commands.append('gshred -u')
            commands.append('sdelete')
            commands.append('eraserl.exe -silent -file {file}')
            commands.append(
                'Eraser.exe addtask --schedule=now -q --file={file}')
        for command in commands:
            try:
                newcommand = command.format(file=quote(fname))
                args = shlex.split(newcommand)
                if command == newcommand:
                    args.append(quote(fname))
                if 0 == check_call(args):
                    getLogger().debug("success running {}".format(command))
                    break
            except Exception:
                getLogger().debug("error running {}".format(command),
                                  exc_info=True)
        if self.still_exists(fname) and insecure is True:
            getLogger().debug("insecurely unlinking {}".format(fname))
            try:
                os.unlink(fname)
            except IOError as e:
                if e.errno != errno.ENOENT:
                    raise

    def shred_archive(self, fname, srm=None, insecure=False):
        fname = os.path.join(self._cache_dir('archives'), fname)
        self.shred_file(fname, srm, insecure)
        return fname

    def shred_cert(self, aid, countdown=[], srm=None, insecure=False):
        path = None
        for fname, docs in self._for_adf('certs').iteritems():
            if 'links' in docs and aid == docs['links'].download:
                path = fname
            elif 'signature' in docs and aid == docs['signature'].aid:
                path = fname
        if path:
            for a in countdown:
                pass
            self.shred_file(os.path.join(self._cache_dir('certs'), path), srm, insecure)
            return path

    def still_exists(self, fname):
        for num in range(3):
            if not os.path.exists(fname):
                return False
            time.sleep(1)
        return os.path.exists(fname)

    def export_cert(self, aid):
        for fname, docs in self._for_adf('certs').iteritems():
            if 'signature' in docs and aid == docs['signature'].aid:
                text = 'longaccess-{}.yaml'.format(aid)
                with open(text, 'w') as f:
                    make_adf(list(docs.itervalues()), out=f, pretty=True)
                return text

    def print_cert(self, aid):
        for fname, docs in self._for_adf('certs').iteritems():
            if 'signature' in docs and aid == docs['signature'].aid:
                html = 'longaccess-{}.html'.format(aid)
                with open(html, 'w') as f:
                    f.write(self._printable_cert(docs))
                return html

    @contains(dict)
    def certs(self, files=[]):
        if not files:
            files = iglob(os.path.join(self._cache_dir('certs'), '*.adf'))
        for f in files:
            with open(f) as fh:
                try:
                    docs = load_archive(fh)
                    if 'signature' in docs:
                        yield (docs['signature'].aid, docs)
                except InvalidArchiveError:
                    getLogger().debug(f, exc_info=True)

    def data_file(self, link):
        try:
            parsed = urlparse(link.local)
            assert not parsed.scheme or parsed.scheme == 'file'
            return os.path.join(self.home, parsed.path)
        except:
            return None

    @property
    def log(self):
        try:
            return os.path.join(self._cache_dir('logs', True), "log.txt")
        except:
            return None

    def merge_prefs(self, prefs):
        try:
            def update(d, u):
                for k, v in u.iteritems():
                    if isinstance(v, collections.Mapping):
                        d[k] = update(d.get(k, {}), v)
                    else:
                        d[k] = u[k]
                return d

            with open(os.path.join(self.home, "preferences.json"), 'r+') as pf:
                return update(prefs, json.load(pf))
        except Exception:
            getLogger().debug("Error reading preferences", exc_info=True)
            return prefs

    def save_prefs(self, prefs):
        with open(os.path.join(self.home, "preferences.json"), 'w') as pf:
            json.dump({"gui": prefs['gui']}, pf)


if __name__ == "__main__":
    import hashlib
    cache = Cache(os.path.expanduser(os.path.join("~", ".longaccess")))
    for fname, docs in cache._for_adf('archives').iteritems():
        path = os.path.join(cache.home, docs['links'].local)
        if not os.path.exists(path):
            continue
        with open(path) as f:
            md5 = hashlib.md5()
            while 1:
                buf = f.read(16*1024)
                if not buf:
                    break
                md5.update(buf)
            assert md5.digest() == docs['auth'].md5, path
# vim: et:sw=4:ts=4



from setuptools import setup, find_packages, Command
from setuptools.command.build_py import build_py as _build_py
from distutils.spawn import find_executable
from lacli import __version__
import os


class gen_thrift(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        thrift = find_executable('thrift')
        if thrift is not None:
            root = os.path.abspath(os.path.dirname(__file__))
            out = os.path.join(root, 'lacli', 'server', 'interface')
            self.mkpath(out)
            self.announce("thrift found, regenerating files in " + out)
            if not self.dry_run:
                self.spawn([thrift, '-out', out,
                            '-v', '--gen', 'py:twisted,new_style',
                            os.path.join(root, 'lacli', 'server', 'ClientInterface.thrift')])
        else:
            self.warn("thrift executable not found, skipping")

try:
    import pandoc
    pandoc.core.PANDOC_PATH = 'pandoc'
    readme = pandoc.Document()
    readme.markdown = open('README.md').read()
    description = readme.rst
except (IOError, ImportError):
    description = 'This is the prototype client program' + \
        'for interacting with the Longaccess service.'


class build_py(_build_py):
    def thrift_gen(self, files):
        pass

    def run(self):
        _build_py.run_command(self, 'gen_thrift')
        _build_py.run(self)


def pep386adapt(version):
    if version is not None and '-' in version:
        # adapt git-describe version to be in line with PEP 386
        parts = version.split('-')
        parts[-2] = 'post'+parts[-2]
        version = '.'.join(parts[:-1])
    return version


setup(version=pep386adapt(__version__),
      name="lacli",
      author="Konstantinos Koukopoulos",
      author_email='kk@longaccess.com',
      description="The Long Access client",
      long_description=description,
      url='http://github.com/longaccess/longaccess-client/',
      license='Apache',
      packages=find_packages(exclude=['features*', '*.t']),
      install_requires=[
          'boto>=2.14, <3',
          'python-dateutil>=2.1, <3',
          'filechunkio>=1.5',
          'docopt>=0.6.1, <0.7',
          'progressbar>=2.3, <3',
          'logutils>=0.3.3, <0.4',
          'requests>=2.0.0, <3',
          'pycrypto>=2.6, <2.7',
          'pyaml>=13.07, <14',
          'Twisted>=13.2, <14',
          'crochet>=1.0.0, <2',
          'treq>=0.2.0, <0.3',
          'pyOpenSSL>=0.13, <1',
          'thrift>=0.9, <1',
          'blessings>=1.5, <2'
      ],
      tests_require=['testtools'],
      test_suite="lacli.t",
      entry_points="""
      [console_scripts]
      lacli = lacli.main:main
      ladec = lacli.cipher.dec:main
      """,
      cmdclass={'build_py': build_py, 'gen_thrift': gen_thrift},
      package_data={'lacli': ['data/certificate.html']},
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Information Technology',
          'Natural Language :: English',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: System :: Archiving',
          'Topic :: Utilities',
      ],
      )



from lacli.exceptions import PauseEvent
from lacli.pool import MPConnection, MPUpload, MPFile
from contextlib import contextmanager
from lacli.log import LogHandler, getLogger
from lacli.control import ControlHandler
from lacli.worker import WorkerPool
from lacli.decorators import block
from twisted.internet import defer, threads
from itertools import count
from multiprocessing import TimeoutError
import json
import errno
import signal


class UploadState(object):
    states = None
    
    @classmethod
    def has_state(cls, fname):
        if cls.states is None:
            cls.setup()
        if fname in cls.states:
            return True
        return False
    
    @classmethod
    def init(cls, cache):
        cls.cache = cache

    @classmethod
    def setup(cls):
        uploads = cls.cache._get_uploads()
        a = cls.cache._for_adf('archives')
        sz = lambda f: a[f]['archive'].meta.size
        cls.states = {k: cls(k, sz(k), **v)
            for k, v in uploads.iteritems()
            if k in a}

    @classmethod
    def get(cls, fname, size=None, capsule=None):
        if cls.states is None:
            cls.setup()
        if fname in cls.states:
            if size is not None:
                cls.states[fname].size = size
            if capsule is not None:
                cls.states[fname].capsule = capsule
            return cls.states[fname]
        cls.states[fname] = UploadState(fname, size, capsule=capsule)
        return cls.states[fname]

    @classmethod
    def reset(cls, fname):
        if fname not in cls.states:
            raise ValueError("Upload doesn't exist!")
        cls.cache._del_upload(fname)
        return cls.states.pop(fname)
    
    def __init__(self, archive, size, uri=None, keys=[], capsule=None, exc=None, paused=True):
        self.cache = type(self).cache
        self.archive = archive
        self.logfile = self.control = None
        self.keys = keys
        self.progress = reduce(lambda x, y: x + y['size'], self.keys, 0)
        self.size = size
        self.pausing = False
        self._paused = paused
        self.uri = uri
        self.exc = exc
        self.capsule = capsule
        self.deferred_upload = None

    def __enter__(self):
        try:
            self.exc = None
            self.control = ControlHandler()
            self.logfile = self.cache._upload_open(self.archive, mode='r+')
            getLogger().debug("Found state file for %s", self.archive)
        except IOError as e:
            if e.errno == errno.ENOENT:
                getLogger().debug("Creating state file for %s", self.archive)
                self.logfile = self.cache._upload_open(self.archive, mode='w+')
            else:
                raise e
        # update keys from file
        upload = self.cache._validate_upload(self.logfile)
        self.uri = upload.get('uri', self.uri)
        self.keys = upload.get('keys', self.keys)
        if self._paused is True:
            self._paused = False
            self.cache._write_upload(
                self.uri, self.capsule, self.logfile,
                self.exc, self._paused)
        return self
    
    @block
    @defer.inlineCallbacks
    def wait_for_upload(self):
        try:
            result = yield self.deferred_upload
            defer.returnValue(result)
        except PauseEvent:
            pass
        if self.pausing is True:
            self.paused()

    def __exit__(self, type, value, traceback):
        if type is not None:
            if type == PauseEvent:
                getLogger().debug("upload paused.")
                self.paused()
            else:
                getLogger().debug("error in upload", exc_info=True)
                self.error(value)
        if self.logfile is not None:
            self.logfile.close()
        self.logfile = self.control = None
        return type is None

    def keydone(self, key, size):
        assert self.logfile is not None, "Log not open"
        self.keys.append(self.cache._checkpoint_upload(key, size, self.logfile))

    def update(self, progress):
        self.progress = progress

    @property
    def seq(self):
        return len(self.keys)

    def pause(self):
        if not self.pausing and self.control is not None:
            self.control.pause()
        self.pausing = True

    def paused(self):
        getLogger().debug("upload state paused")
        if self.exc is not None:
            getLogger().debug("can't pause a failed upload")
            return
        self._paused = True
        if self.pausing is True:
            self.pausing = False
        self.cache._write_upload(
            self.uri, self.capsule, self.logfile,
            self.exc, self._paused)

    def active(self):
        return self._paused is False and self.exc is None

    def signal(self, sig, frame):
        getLogger().debug("Got interrupt")
        if sig == signal.SIGINT:
            getLogger().debug("Pausing")
            if self.pausing is True:
                raise SystemExit("Interrupted")
            self.pausing = True
            self.control.pause()
            
    def save_op(self, op):
        assert self.uri is None, "Can't change URI for upload state"
        if op.uri is None:
            return
        self.cache._write_upload(op.uri, self.capsule, self.logfile, self.exc, self._paused)
        self.uri = op.uri

    def error(self, exc):
        if self.exc is None:
            self.cache._write_upload(self.uri, self.capsule, self.logfile, str(exc), self._paused)
            self.exc = exc


class Upload(object):
    def __init__(self, session, nprocs, debug, state):
        self.prefs = {
            'nprocs': nprocs,
            'debugworker': debug > 2
        }
        self.log = LogHandler()
        self.state = state

    @contextmanager
    def _workers(self, progq):
        with self.log as logq:
            with self.state.control as ctrlq:
                pool = WorkerPool(
                    self.prefs, logq, progq, ctrlq)
                try: 
                    yield pool
                finally:
                    getLogger().debug("terminating pool")
                    pool.terminate()
                    pool.join()

    def upload_temp(self, token, source, etags, pool, seq):
        key = "{prefix}temp-archive-{seq}".format(
            prefix=token['prefix'], seq=seq)
        connection = MPConnection(token)
        with MPUpload(connection, source, key) as uploader:
            etags[key], source = uploader.get_result(
                uploader.submit_job(pool))
        return source

    @defer.inlineCallbacks
    def upload(self, fname, upload, progq):
        with self._workers(progq) as pool:
            etags = {}
            source = MPFile(fname, self.state.progress)

            for seq in count(start=self.state.seq):
                token = yield upload.status
                try:
                    source = yield threads.deferToThread(
                        self.upload_temp, token, source, etags, pool, seq)
                except PauseEvent:
                    getLogger().debug("paused after uploading %d temporary keys", seq)
                    raise
                except TimeoutError:
                    getLogger().debug("timeout after uploading %d temporary keys", seq)
                if source is None:
                    getLogger().debug("uploaded entire archive")
                    break

            getLogger().debug("uploaded %d temp keys", len(etags))
            for key, tag in etags.iteritems():
                getLogger().debug("key: %s (etag: %s)", key, tag)
# vim: et:sw=4:ts=4



import sys
import os
import signal

from lacli.log import logToQueue, getLogger
from lacli.progress import progressToQueue
from lacli.control import controlByQueue
from multiprocessing import cpu_count, pool, current_process, Process
try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = lambda x: x


def initworker(logq, progq, ctrlq, stdin=None):
    """initializer that sets up logging and progress from sub procs """
    logToQueue(logq)
    progressToQueue(progq)
    controlByQueue(ctrlq)
    setproctitle(current_process().name)
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if hasattr(signal, 'SIGBREAK'):
        signal.signal(signal.SIGBREAK, signal.SIG_IGN)
    getLogger().debug("Worker " + current_process().name + " logging started")
    if stdin is not None:
        sys.stdin = os.fdopen(stdin, 'r')
        print "Worker " + current_process().name + " opened stdin"


class WorkerProcess(Process):
    def __init__(self, *args, **kwargs):
        super(WorkerProcess, self).__init__(*args, **kwargs)
        self.name = "Long Access " + self.name


class WorkerPool(pool.Pool):
    Process = WorkerProcess

    def __init__(self, prefs, logq, progq, ctrlq):
        nprocs = prefs['nprocs'] or max(cpu_count()-1, 1)
        args = [logq, progq, ctrlq]
        if prefs['debugworker']:
            args.append(sys.stdin.fileno())
        super(WorkerPool, self).__init__(nprocs, initworker, args)
        getLogger().debug("set up pool of {} procs..".format(nprocs))
# vim: et:sw=4:ts=4



from testtools import TestCase
from mock import Mock


class UploadTest(TestCase):
    def setUp(self):
        super(UploadTest, self).setUp()

    def tearDown(self):
        super(UploadTest, self).tearDown()

    def _makeit(self, *args, **kw):
        from lacli.upload import Upload
        return Upload(*args,  **kw)

    def test_upload(self):
        assert self._makeit(Mock(), 4, 4)
